# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# -------------------------------------------------------------------------
import multiprocessing as mp
import os


def get_start_method():
    return os.getenv('TORCH_PROFILER_START_METHOD', 'spawn')


__all__ = [x for x in dir(mp.get_context(get_start_method())) if not x.startswith('_')]
globals().update((name, getattr(mp.get_context(get_start_method()), name)) for name in __all__)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# -------------------------------------------------------------------------
from typing import List, Tuple


# src_ranges: item of (start_time, end_time, value)
def merge_ranges_with_value(src_ranges):
    from collections import namedtuple
    from enum import IntEnum

    class EndpointTypes(IntEnum):
        START = 0
        END = 1

    EndPoint = namedtuple('EndPoint', ['time', 'pt_type', 'value'])

    merged_ranges = []
    if len(src_ranges) > 0:
        # Build tuple of (time, type, value)
        endpoints: List[EndPoint] = []
        for r in src_ranges:
            endpoints.append(EndPoint(r[0], EndpointTypes.START, r[2]))
            endpoints.append(EndPoint(r[1], EndpointTypes.END, r[2]))
        endpoints.sort(key=lambda x: [x.time, int(x.pt_type)])  # Make START in front of END if equal on time.

        last_endpoint_time = endpoints[0].time
        last_value = endpoints[0].value
        for i in range(1, len(endpoints)):
            ep = endpoints[i]
            if ep.time > last_endpoint_time and last_value > 0.0:
                approximated_sm_efficiency = min(last_value, 1.0)
                merged_ranges.append((last_endpoint_time, ep.time, approximated_sm_efficiency))
            last_endpoint_time = ep.time
            if ep.pt_type == EndpointTypes.START:
                last_value += ep.value
            else:
                last_value -= ep.value

    return merged_ranges


# range_list1 item is length 3. range_list2 item is length 2.
# Reture value's item is length 3.
def intersection_ranges_lists_with_value(range_list1, range_list2) -> List[Tuple[int, int, int]]:
    range_list_dst = []
    if len(range_list1) == 0 or len(range_list2) == 0:
        return range_list_dst
    r1 = range_list1[0]
    r2 = range_list2[0]
    i1 = i2 = 0
    while i1 < len(range_list1):
        if i2 == len(range_list2):
            break
        elif r2[1] <= r1[0]:
            r2, i2 = pop_list(range_list2, i2)
        elif r2[0] <= r1[0] and r2[1] < r1[1]:
            assert (r2[1] > r1[0])
            range_list_dst.append((r1[0], r2[1], r1[2]))
            r1 = (r2[1], r1[1], r1[2])
            r2, i2 = pop_list(range_list2, i2)
        elif r2[0] <= r1[0]:
            assert (r2[1] >= r1[1])
            range_list_dst.append(r1)
            r2 = (r1[1], r2[1])
            r1, i1 = pop_list(range_list1, i1)
        elif r2[1] < r1[1]:
            assert (r2[0] > r1[0])
            range_list_dst.append((r2[0], r2[1], r1[2]))
            r1 = (r2[1], r1[1], r1[2])
            r2, i2 = pop_list(range_list2, i2)
        elif r2[0] < r1[1]:
            assert (r2[1] >= r1[1])
            range_list_dst.append((r2[0], r1[1], r1[2]))
            r2 = (r1[1], r2[1])
            r1, i1 = pop_list(range_list1, i1)
        else:
            assert (r2[0] >= r1[1])
            r1, i1 = pop_list(range_list1, i1)
    return range_list_dst


def subtract_ranges_lists(range_list1: List[Tuple[int, int]],
                          range_list2: List[Tuple[int, int]]) -> List[Tuple[int, int]]:
    range_list_dst = []
    if len(range_list1) == 0:
        return range_list_dst
    if len(range_list2) == 0:
        range_list_dst = list(range_list1)
        return range_list_dst
    r1 = range_list1[0]
    r2 = range_list2[0]
    i1 = i2 = 0
    while i1 < len(range_list1):
        if i2 == len(range_list2):
            range_list_dst.append(r1)
            r1, i1 = pop_list(range_list1, i1)
        elif r2[1] <= r1[0]:
            r2, i2 = pop_list(range_list2, i2)
        elif r2[0] <= r1[0] and r2[1] < r1[1]:
            r1 = (r2[1], r1[1])
            r2, i2 = pop_list(range_list2, i2)
        elif r2[0] <= r1[0]:
            assert (r2[1] >= r1[1])
            r2 = (r1[1], r2[1])
            r1, i1 = pop_list(range_list1, i1)
        elif r2[0] < r1[1]:
            assert (r2[0] > r1[0])
            range_list_dst.append((r1[0], r2[0]))
            r1 = (r2[0], r1[1])
        else:
            assert (r2[0] >= r1[1])
            range_list_dst.append(r1)
            r1, i1 = pop_list(range_list1, i1)
    return range_list_dst


def intersection_ranges_lists(range_list1: List[Tuple[int, int]],
                              range_list2: List[Tuple[int, int]]) -> List[Tuple[int, int]]:
    range_list_dst = []
    if len(range_list1) == 0 or len(range_list2) == 0:
        return range_list_dst
    r1 = range_list1[0]
    r2 = range_list2[0]
    i1 = i2 = 0
    while i1 < len(range_list1):
        if i2 == len(range_list2):
            break
        elif r2[1] <= r1[0]:
            r2, i2 = pop_list(range_list2, i2)
        elif r2[0] <= r1[0] and r2[1] < r1[1]:
            assert (r2[1] > r1[0])
            range_list_dst.append((r1[0], r2[1]))
            r1 = (r2[1], r1[1])
            r2, i2 = pop_list(range_list2, i2)
        elif r2[0] <= r1[0]:
            assert (r2[1] >= r1[1])
            range_list_dst.append(r1)
            r2 = (r1[1], r2[1])
            r1, i1 = pop_list(range_list1, i1)
        elif r2[1] < r1[1]:
            assert (r2[0] > r1[0])
            range_list_dst.append(r2)
            r1 = (r2[1], r1[1])
            r2, i2 = pop_list(range_list2, i2)
        elif r2[0] < r1[1]:
            assert (r2[1] >= r1[1])
            range_list_dst.append((r2[0], r1[1]))
            r2 = (r1[1], r2[1])
            r1, i1 = pop_list(range_list1, i1)
        else:
            assert (r2[0] >= r1[1])
            r1, i1 = pop_list(range_list1, i1)
    return range_list_dst


def get_ranges_sum(ranges: List[Tuple[int, int]]) -> int:
    sum: int = 0
    for range in ranges:
        sum += (range[1] - range[0])
    return sum


def pop_list(range_list, index):
    next_index = index + 1
    if next_index >= len(range_list):
        return None, len(range_list)
    next_item = range_list[next_index]
    return next_item, next_index


def merge_ranges(src_ranges, is_sorted=False) -> List[Tuple[int, int]]:
    if not src_ranges:
        # return empty list if src_ranges is None or its length is zero.
        return []

    if not is_sorted:
        src_ranges.sort(key=lambda x: x[0])

    merged_ranges = []
    merged_ranges.append(src_ranges[0])
    for src_id in range(1, len(src_ranges)):
        src_range = src_ranges[src_id]
        if src_range[1] > merged_ranges[-1][1]:
            if src_range[0] <= merged_ranges[-1][1]:
                merged_ranges[-1] = (merged_ranges[-1][0], src_range[1])
            else:
                merged_ranges.append((src_range[0], src_range[1]))

    return merged_ranges



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# -------------------------------------------------------------------------
from google.cloud import storage

from .. import utils
from .base import BaseFileSystem, RemotePath, StatData

logger = utils.get_logger()


class GoogleBlobSystem(RemotePath, BaseFileSystem):
    """Provides filesystem access to S3."""

    def __init__(self):
        if not storage:
            raise ImportError('google-cloud-storage must be installed for Google Cloud Blob support.')

    def exists(self, dirname):
        """Returns whether the path is a directory or not."""
        bucket_name, path = self.bucket_and_path(dirname)
        client = self.create_google_cloud_client()
        bucket = client.bucket(bucket_name)
        return bucket.blob(path).exists()

    def read(self, filename, binary_mode=False, size=None, continue_from=None):
        raise NotImplementedError

    def write(self, filename, file_content, binary_mode=False):
        raise NotImplementedError

    def glob(self, filename):
        raise NotImplementedError

    def download_file(self, file_to_download, file_to_save):
        bucket_name, path = self.bucket_and_path(file_to_download)
        client = self.create_google_cloud_client()
        bucket = client.bucket(bucket_name)
        blob = bucket.blob(path)
        blob.download_to_filename(file_to_save)

    def isdir(self, dirname):
        """Returns whether the path is a directory or not."""
        basename, parts = self.split_blob_path(dirname)
        if basename is None or parts is None:
            return False
        if basename == '':
            # root container case
            return True
        else:
            return basename == parts[0] and len(parts) > 1

    def listdir(self, dirname):
        """Returns a list of entries contained within a directory."""
        bucket_name, path = self.bucket_and_path(dirname)
        client = self.create_google_cloud_client()
        blobs = client.list_blobs(bucket_name, prefix=path)
        items = []
        for blob in blobs:
            item = self.relpath(blob.name, path)
            if items not in items:
                items.append(item)
        return items

    def makedirs(self, dirname):
        """No need create directory since the upload blob will automatically create"""
        pass

    def stat(self, filename):
        """Returns file statistics for a given path."""
        bucket_name, path = self.bucket_and_path(filename)
        client = self.create_google_cloud_client()
        bucket = client.bucket(bucket_name)
        blob = bucket.get_blob(path)
        return StatData(blob.size)

    def walk(self, top, topdown=True, onerror=None):
        bucket_name, path = self.bucket_and_path(top)
        client = self.create_google_cloud_client()
        blobs = client.list_blobs(bucket_name, prefix=path)
        results = {}
        for blob in blobs:
            dirname, basename = self.split(blob.name)
            dirname = 'gs://{}/{}'.format(bucket_name, dirname)
            results.setdefault(dirname, []).append(basename)
        for key, value in results.items():
            yield key, None, value

    def split_blob_path(self, blob_path):
        """ Find the first blob start with blob_path, then get the relative path starting from dirname(blob_path).
        Finally, split the relative path.
        return (basename(blob_path), [relative splitted paths])
        If blob_path doesn't exist, return (None, None)
        For example,
            For blob gs://tests/test1/test2/test.txt
            * If the blob_path is '', return ('', [test1, test2, test.txt])
            * If the blob_path is test1, return (test1, [test2, test.txt])
            * If the blob_path is test1/test2, return (test2, [test2, test.txt])
            * If the blob_path is test1/test2/test.txt, return (test.txt, [test.txt])
        """
        bucket_name, path = self.bucket_and_path(blob_path)
        client = self.create_google_cloud_client()
        blobs = client.list_blobs(bucket_name, prefix=path, delimiter=None, max_results=1)

        for blob in blobs:
            dir_path, basename = self.split(path)
            if dir_path:
                rel_path = blob.name[len(dir_path):]
                parts = rel_path.lstrip('/').split('/')
            else:
                parts = blob.name.split('/')
            return (basename, parts)
        return (None, None)

    def bucket_and_path(self, url):
        """Split an S3-prefixed URL into bucket and path."""
        if url.startswith('gs://'):
            url = url[len('gs://'):]
        idx = url.index('/')
        bucket = url[:idx]
        path = url[(idx + 1):]
        return bucket, path

    def create_google_cloud_client(self):
        # TODO: support client with credential?
        client = storage.Client.create_anonymous_client()
        return client



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# -------------------------------------------------------------------------
from abc import ABCMeta
from typing import List, Tuple, Union

from ..node import DeviceNode, OperatorNode
from ..op_agg import aggregate_ops
from .contract import OpAgg


class Operator(metaclass=ABCMeta):
    def __init__(self, name) -> None:
        self.name: str = name

    def __str__(self) -> str:
        return f'{self.name}: {self.duration}'

    @property
    def duration(self) -> int:
        return 0

    @property
    def device_duration(self) -> int:
        return 0

    @property
    def total_duration(self):
        return self.device_duration or self.duration

    def aggregate_ops(self):
        ops, _ = self.get_operators_and_kernels()
        agg_result = aggregate_ops(ops, [lambda x: x.name])[0]
        for agg in agg_result.values():
            yield OpAgg(
                agg.name,
                agg.calls,
                agg.host_duration,
                agg.device_duration,
                agg.self_host_duration,
                agg.self_device_duration)

    def get_operators_and_kernels(self) -> Tuple[List[OperatorNode], List[DeviceNode]]:
        return [], []


class BlankOp(Operator):
    def __init__(self) -> None:
        super().__init__('Blank')


class UnknownOp(Operator):
    def __init__(self, device_duration: int, duration: int) -> None:
        super().__init__('Unknown')
        self.device_duration = device_duration
        self.duration = duration

    @property
    def duration(self) -> int:
        return self.duration

    @property
    def device_duration(self) -> int:
        return self.device_duration


class Operators(Operator):
    def __init__(self, nodes: Union[OperatorNode, List[OperatorNode]]):
        if not nodes:
            raise ValueError('the operator node is None or empty')
        if isinstance(nodes, OperatorNode):
            super().__init__(nodes.name)
        elif isinstance(nodes, list):
            super().__init__('CompositeNodes')

        self.op_nodes: Union[OperatorNode, List[OperatorNode]] = nodes

    @property
    def duration(self):
        if isinstance(self.op_nodes, list):
            return sum(n.duration for n in self.op_nodes)
        else:
            return self.op_nodes.duration

    @property
    def device_duration(self):
        if isinstance(self.op_nodes, list):
            return sum(n.device_duration for n in self.op_nodes)
        else:
            return self.op_nodes.device_duration

    @property
    def total_duration(self):
        if isinstance(self.op_nodes, list):
            return sum(n.device_duration or n.duration for n in self.op_nodes)
        else:
            return self.op_nodes.device_duration or self.op_nodes.duration

    def __str__(self) -> str:
        if isinstance(self.op_nodes, list):
            return f'{self.name}: {len(self.op_nodes)}: {self.op_nodes[0].name}: {self.total_duration}'
        else:
            return f'{self.name}: {self.op_nodes.__class__.__name__}: {self.total_duration}'

    def get_operators_and_kernels(self) -> Tuple[List[OperatorNode], List[DeviceNode]]:
        if isinstance(self.op_nodes, list):
            nodes = self.op_nodes
        else:
            nodes = [self.op_nodes]

        ops: List[OperatorNode] = []
        kernels: List[DeviceNode] = []
        for n in nodes:
            o, k = n.get_operator_and_kernels()
            ops.extend(o)
            kernels.extend(k)
        return ops, kernels


def create_operator(op_nodes: Union[OperatorNode, List[OperatorNode]]) -> Operator:
    if op_nodes:
        return Operators(op_nodes)
    else:
        return BlankOp()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# -------------------------------------------------------------------------
from collections import namedtuple
from typing import Dict, Generator, Iterable, List, Optional, Set, Tuple, Union

from .node import (DataLoaderNode, ModuleNode, OperatorNode, OptimizerNode,
                   PLModuleNode, ProfilerStepNode, is_operator_node)
from .trace import BaseEvent, EventTypes, PLModuleEvent, PythonFunctionEvent


class Module:
    def __init__(self, name: str, module_id: int, shape: str = ''):
        self.name = name
        self.module_id = module_id
        self.children: List[Module] = []

    def __hash__(self):
        return hash((self.name, self.module_id, tuple(self.children)))

    def __eq__(self, o) -> bool:
        if not isinstance(o, Module):
            return False

        return (self.name == o.name and
                self.module_id == o.module_id and
                self.children == o.children)


class ModuleStats:
    def __init__(self, name: str, module_id: int):
        self.name = name
        self.module_id = module_id
        self.occurences: int = 0
        self.operators: int = 0
        self.host_duration: int = 0
        self.device_duration: int = 0
        self.self_host_duration: int = 0
        self.self_device_duration: int = 0

    @property
    def avg_host_duration(self):
        return self.host_duration / self.occurences

    @property
    def avg_device_duration(self):
        return self.device_duration / self.occurences


Stats = namedtuple('Stats', [
    'name',
    'id',
    'occurences',
    'operators',
    'host_duration',
    'self_host_duration',
    'device_duration',
    'self_device_duration',
    'avg_duration',
    'children'])


def aggegate_module_view(tid2tree: Dict[int, OperatorNode], events: List[BaseEvent]) -> Optional[List[Stats]]:
    roots = _build_module_hierarchy(events)
    modules = _get_node_list(tid2tree, ModuleNode)
    if modules and roots:
        return _process_module_statistics(modules, roots)
    else:
        return None


def aggegate_pl_module_view(tid2tree: Dict[int, OperatorNode], events: List[BaseEvent]) -> Optional[List[Stats]]:
    roots = _build_module_hierarchy_from_name(events)
    modules = _get_node_list(tid2tree, PLModuleNode)
    if modules and roots:
        return _process_module_statistics(modules, roots)
    else:
        return None


def _build_module_hierarchy_from_name(events: List[PLModuleEvent]) -> List[Module]:
    pl_module_events = [e for e in events if e.type == EventTypes.PL_MODULE]
    name2module: Dict[str, Module] = {}
    no_root: Set[str] = set()

    for event in pl_module_events:
        if event.name not in name2module:
            name2module[event.name] = Module(event.name, 0)

    for name, module in name2module.items():
        if name.find('.') == -1:
            continue
        parent_name = name[:name.rfind('.')]
        if parent_name in name2module:
            name2module[parent_name].children.append(module)
            no_root.add(module.name)

    return [module for name, module in name2module.items() if name not in no_root]


def _build_module_hierarchy(events: List[PythonFunctionEvent]) -> List[Module]:
    """Get the module hierarchy from the chome trace events
    """
    python_events = [e for e in events if e.type in (EventTypes.PYTHON_FUNCTION, EventTypes.MODULE)]
    id_to_event = {e.python_id: e for e in python_events}

    # Extract Python function topology.
    children: Dict[int, List[int]] = {}
    for e in python_events:
        e_id = e.python_id
        children.setdefault(e_id, [])
        e_parent_id = e.python_parent_id
        children.setdefault(e_parent_id, [])
        children[e_parent_id].append(e_id)
    function_leaves = [k for k, v in children.items() if not v]

    # Convert Python function topology to Module topology.
    #   This is a simple O(n) tree walking algorithm where we start from the leaves
    #   and walk up, discarding any nodes which are not Module nodes.
    module_parent_map = {}
    seen = set()
    for i in function_leaves:
        e = id_to_event[i]
        current_module = None
        while e is not None:
            e_id = e.python_id
            if e.type == EventTypes.MODULE:
                if current_module is not None:
                    module_parent_map[current_module.python_id] = e_id
                current_module = e
                module_parent_map.setdefault(e_id, None)

            seen_key = (e_id, id(current_module))
            if seen_key in seen:
                break
            seen.add(seen_key)

            e = id_to_event.get(e.python_parent_id, None)

    module_roots = [k for k, v in module_parent_map.items() if v is None]
    module_child_map: Dict[int, List[int]] = {}
    for child_id, parent_id in module_parent_map.items():
        module_child_map.setdefault(child_id, [])
        module_child_map.setdefault(parent_id, [])
        module_child_map[parent_id].append(child_id)

    # The traverse order is well defined which guarantees that a given topology
    # will produce a unique and unambiguous hierarchy.
    def append_hierarchy(e_id) -> Module:
        e = id_to_event[e_id]
        module = Module(e.name, e.module_id)
        for id in module_child_map[e_id]:
            child = append_hierarchy(id)
            module.children.append(child)
        return module

    unique_modules: Set[Module] = set()
    for e_id in module_roots:
        root = append_hierarchy(e_id)
        unique_modules.add(root)

    return list(unique_modules)


def _aggregate_modules(modules: Iterable[Union[ModuleNode, PLModuleNode]]) -> Dict[Tuple[str, int], ModuleStats]:
    """Aggregate the modules based on the name and module_id"""
    module_aggs: Dict[Tuple(str, int), ModuleStats] = {}
    for m in modules:
        key = (m.name, m.module_id)
        if key not in module_aggs:
            module_aggs[key] = ModuleStats(m.name, m.module_id)
        agg = module_aggs[key]
        agg.occurences += 1

        agg.operators += sum(is_operator_node(child) for child in m.children)

        agg.self_host_duration += m.self_host_duration
        agg.host_duration += m.end_time - m.start_time

        agg.self_device_duration += m.self_device_duration
        agg.device_duration += m.device_duration

    return module_aggs


def _get_node_list(tid2tree: Dict[int, OperatorNode], node_class) -> Generator[OperatorNode, None, None]:
    """Get all node with node_class from the operator tree"""
    def traverse_node(node):
        # Check OptimizerNode here because in PytorchLightning PLModuleNode is under OptimizerNoder.
        if type(node) not in (ProfilerStepNode, ModuleNode, OperatorNode, OptimizerNode, PLModuleNode, DataLoaderNode):
            return

        if isinstance(node, node_class):
            yield node

        for child in node.children:
            yield from traverse_node(child)

    for _, root in tid2tree.items():
        for child in root.children:
            yield from traverse_node(child)


def _process_module_statistics(
        modules_nodes: Iterable[Union[ModuleNode, PLModuleNode]],
        hierarchy: Iterable[Module]) -> List[Stats]:
    """Get the module statistics from the ModuleNode(s) and the hierarchy
    """
    module_aggs = _aggregate_modules(modules_nodes)

    def process_modules(h_modules: Iterable[Module]):
        for m in h_modules:
            name = m.name.replace('nn.Module: ', '')
            stats = module_aggs[(m.name, m.module_id)]

            child_stats = list(process_modules(m.children))
            yield Stats(
                name,
                m.module_id,
                stats.occurences,
                stats.operators,
                stats.host_duration,
                stats.self_host_duration,
                stats.device_duration,
                stats.self_device_duration,
                stats.avg_device_duration if stats.avg_device_duration > 0 else stats.avg_host_duration,
                child_stats)

    data = sorted(process_modules(hierarchy), key=lambda x: x.name)
    return data


def get_module_tree(tid2tree: Dict[int, OperatorNode]):
    """Get the module tree in timeline"""
    from copy import copy

    modules = []

    def traverse_node(node, parent: Optional[ModuleNode]):
        if type(node) not in (ProfilerStepNode, ModuleNode):
            return

        if isinstance(node, ModuleNode):
            module = copy(node)
            # remove the children after copy to keep the module only
            module.children = []

            if parent is None:
                modules.append(module)
            else:
                parent.children.append(module)
            parent = module

        for child in node.children:
            traverse_node(child, parent)

    for _, root in tid2tree.items():
        for child in root.children:
            # since the root node is CallTreeRoot, there is no parent ModuleNode
            traverse_node(child, None)

    return modules


def dump_modules(level: int, modules: Iterable[Union[Module, ModuleNode]]):
    """testing purpose"""
    for module in modules:
        print(f"{'    ' * level}{module.name.replace('nn.Module: ', '')}_{module.module_id}")
        dump_modules(level + 1, module.children)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# --------------------------------------------------------------------------
from collections import OrderedDict
from typing import Dict, Iterable, List

from .. import consts, utils
from ..run import DistributedRunProfile, RunProfile
from .data import DistributedRunProfileData, RunProfileData
from .module_op import aggegate_module_view, aggegate_pl_module_view
from .op_agg import KernelAggByNameOp, OperatorAgg
from .overall_parser import ProfileRole

logger = utils.get_logger()


class RunGenerator(object):
    def __init__(self, worker, span, profile_data: RunProfileData):
        self.worker = worker
        self.span = span
        self.profile_data = profile_data

    def generate_run_profile(self):
        profile_run = RunProfile(self.worker, self.span)
        profile_run.is_pytorch_lightning = self.profile_data.is_pytorch_lightning
        profile_run.has_runtime = self.profile_data.has_runtime
        profile_run.has_kernel = self.profile_data.has_kernel
        profile_run.has_communication = self.profile_data.has_communication
        profile_run.has_memcpy_or_memset = self.profile_data.has_memcpy_or_memset
        profile_run.profiler_start_ts = self.profile_data.profiler_start_ts
        profile_run.views.append(consts.OVERALL_VIEW)
        profile_run.overview = self._generate_overview()

        profile_run.views.append(consts.OP_VIEW)
        profile_run.operation_pie_by_name = self._generate_op_pie()
        profile_run.operation_table_by_name = self._generate_op_table(self.profile_data.op_list_groupby_name)
        profile_run.operation_stack_by_name = self._generate_op_table_for_stack(False)
        profile_run.operation_pie_by_name_input = self._generate_op_pie(True)
        profile_run.operation_table_by_name_input = self._generate_op_table(
            self.profile_data.op_list_groupby_name_input, True)
        profile_run.operation_stack_by_name_input = self._generate_op_table_for_stack(True)

        if self.profile_data.has_kernel:
            profile_run.views.append(consts.KERNEL_VIEW)
            profile_run.kernel_op_table = self._generate_kernel_op_table()
            profile_run.kernel_pie = self._generate_kernel_pie()
            profile_run.kernel_table = self._generate_kernel_table()
            profile_run.tc_pie = self._generate_tc_pie()

        profile_run.views.append(consts.TRACE_VIEW)
        profile_run.trace_file_path = self.profile_data.trace_file_path

        profile_run.gpu_metrics = self.profile_data.gpu_metrics_parser.get_gpu_metrics()

        gpu_infos = {gpu_id: RunGenerator._get_gpu_info(self.profile_data.device_props, gpu_id)
                     for gpu_id in self.profile_data.gpu_metrics_parser.gpu_ids}
        gpu_infos = {gpu_id: gpu_info for gpu_id, gpu_info in gpu_infos.items() if gpu_info is not None}

        profile_run.gpu_summary, profile_run.gpu_tooltip = \
            self.profile_data.gpu_metrics_parser.get_gpu_metrics_data_tooltip(
                gpu_infos, self.profile_data.tc_ratio)

        profile_run.tid2tree = self.profile_data.tid2tree
        profile_run.pl_tid2tree = self.profile_data.pl_tid2tree

        if self.profile_data.memory_snapshot:
            profile_run.views.append(consts.MEMORY_VIEW)
            profile_run.memory_snapshot = self.profile_data.memory_snapshot

        profile_run.module_stats = aggegate_module_view(self.profile_data.tid2tree, self.profile_data.events)
        profile_run.pl_module_stats = aggegate_pl_module_view(self.profile_data.tid2tree, self.profile_data.events)
        if profile_run.is_pytorch_lightning and profile_run.pl_module_stats:
            profile_run.views.append(consts.LIGHTNING_VIEW)
        elif profile_run.module_stats:
            profile_run.views.append(consts.MODULE_VIEW)

        return profile_run

    def _generate_overview(self):
        def build_part_time_str(part_cost: float, part_name: str):
            format_str = ('<div class="visualization-tooltip" style="white-space: nowrap;">'
                          'Step {}<br>'
                          'Total: {}us<br>'
                          '<b>{}: {}us</b><br>'
                          'Percentage: {}%'
                          '</div>')
            percentage = round(100 * part_cost / costs.costs[ProfileRole.Total], 2)
            return format_str.format(step_name, costs.costs[ProfileRole.Total], part_name, part_cost, percentage)

        def build_avg_cost_dict(part_name: str, part_cost: float):
            cost_dict = {'name': part_name,
                         'description': '',
                         'value': round(part_cost),
                         'extra': round(100 * part_cost / self.profile_data.avg_costs.costs[ProfileRole.Total], 2)}
            return cost_dict

        show_gpu = (self.profile_data.has_runtime
                    or self.profile_data.has_kernel or self.profile_data.has_memcpy_or_memset)

        column_tootip = {'type': 'string', 'role': 'tooltip', 'p': {'html': 'true'}}
        data = {}
        data['steps'] = {}
        data['steps']['columns'] = [{'type': 'string', 'name': 'Step'}]
        if show_gpu:
            data['steps']['columns'].extend([{'type': 'number', 'name': 'Kernel'},
                                             column_tootip,
                                             {'type': 'number', 'name': 'Memcpy'},
                                             column_tootip,
                                             {'type': 'number', 'name': 'Memset'},
                                             column_tootip])
        if self.profile_data.has_communication:
            data['steps']['columns'].extend([{'type': 'number', 'name': 'Communication'},
                                             column_tootip])
        if show_gpu:
            data['steps']['columns'].extend([{'type': 'number', 'name': 'Runtime'},
                                             column_tootip])
        data['steps']['columns'].extend([{'type': 'number', 'name': 'DataLoader'},
                                         column_tootip,
                                         {'type': 'number', 'name': 'CPU Exec'},
                                         column_tootip,
                                         {'type': 'number', 'name': 'Other'},
                                         column_tootip])

        data['steps']['rows'] = []
        for i in range(len(self.profile_data.steps_costs)):
            costs = self.profile_data.steps_costs[i]
            step_name = self.profile_data.steps_names[i]
            row = [step_name]
            if show_gpu:
                row.extend([costs.costs[ProfileRole.Kernel],
                            build_part_time_str(costs.costs[ProfileRole.Kernel], 'Kernel'),
                            costs.costs[ProfileRole.Memcpy],
                            build_part_time_str(costs.costs[ProfileRole.Memcpy], 'Memcpy'),
                            costs.costs[ProfileRole.Memset],
                            build_part_time_str(costs.costs[ProfileRole.Memset], 'Memset')])
            if self.profile_data.has_communication:
                row.extend([costs.costs[ProfileRole.Communication],
                            build_part_time_str(costs.costs[ProfileRole.Communication], 'Communication')])
            if show_gpu:
                row.extend([costs.costs[ProfileRole.Runtime],
                            build_part_time_str(costs.costs[ProfileRole.Runtime], 'Runtime')])
            row.extend([costs.costs[ProfileRole.DataLoader],
                        build_part_time_str(costs.costs[ProfileRole.DataLoader], 'DataLoader'),
                        costs.costs[ProfileRole.CpuOp],
                        build_part_time_str(costs.costs[ProfileRole.CpuOp], 'CPU Exec'),
                        costs.costs[ProfileRole.Other],
                        build_part_time_str(costs.costs[ProfileRole.Other], 'Other')])
            data['steps']['rows'].append(row)

        avg_costs = []
        if show_gpu:
            avg_costs.extend([
                build_avg_cost_dict('Kernel', self.profile_data.avg_costs.costs[ProfileRole.Kernel]),
                build_avg_cost_dict('Memcpy', self.profile_data.avg_costs.costs[ProfileRole.Memcpy]),
                build_avg_cost_dict('Memset', self.profile_data.avg_costs.costs[ProfileRole.Memset])
            ])
        if self.profile_data.has_communication:
            avg_costs.extend([
                build_avg_cost_dict('Communication', self.profile_data.avg_costs.costs[ProfileRole.Communication])
            ])
        if show_gpu:
            avg_costs.extend([
                build_avg_cost_dict('Runtime', self.profile_data.avg_costs.costs[ProfileRole.Runtime])
            ])
        avg_costs.extend([
            build_avg_cost_dict('DataLoader', self.profile_data.avg_costs.costs[ProfileRole.DataLoader]),
            build_avg_cost_dict('CPU Exec', self.profile_data.avg_costs.costs[ProfileRole.CpuOp]),
            build_avg_cost_dict('Other', self.profile_data.avg_costs.costs[ProfileRole.Other])
        ])

        data['performance'] = [{'name': 'Average Step Time', 'description': '',
                                'value': round(self.profile_data.avg_costs.costs[ProfileRole.Total]),
                                'extra': 100, 'children': avg_costs}]

        if len(self.profile_data.recommendations) == 0:
            html = '<li>N/A</li>'
        else:
            html = ''
            for recommendation in self.profile_data.recommendations:
                html += '<li>{}</li>'.format(recommendation)
        data['recommendations'] = '<ul>{}</ul>'.format(html)

        return data

    def _generate_op_pie(self, group_by_input_shape: bool = False):
        op_device_total_time = []
        op_device_self_time = []
        op_host_total_time = []
        op_host_self_time = []

        if group_by_input_shape:
            op_list = self.profile_data.op_list_groupby_name_input
        else:
            op_list = self.profile_data.op_list_groupby_name

        for op_agg in op_list:
            # Whether device_duration & self_device_duration are accurate or not depends on the input tracing data.
            if op_agg.device_duration > 0:
                op_device_total_time.append([op_agg.name, op_agg.device_duration])
            if op_agg.self_device_duration > 0:
                op_device_self_time.append([op_agg.name, op_agg.self_device_duration])
            if op_agg.host_duration > 0:
                op_host_total_time.append([op_agg.name, op_agg.host_duration])
            if op_agg.self_host_duration > 0:
                op_host_self_time.append([op_agg.name, op_agg.self_host_duration])

        op_device_total_time.sort(key=lambda x: x[1], reverse=True)
        op_device_self_time.sort(key=lambda x: x[1], reverse=True)
        op_host_total_time.sort(key=lambda x: x[1], reverse=True)
        op_host_self_time.sort(key=lambda x: x[1], reverse=True)

        data = {}
        device_total_time = {}
        device_self_time = {}
        host_total_time = {}
        host_self_time = {}

        if len(op_device_total_time) > 0:
            device_total_time['title'] = 'Device Total Time (us)'
            device_total_time['columns'] = [{'type': 'string', 'name': 'name'}, {'type': 'number', 'name': 'value'}]
            device_total_time['rows'] = op_device_total_time
        else:
            device_total_time = None

        if len(op_device_self_time) > 0:
            device_self_time['title'] = 'Device Self Time (us)'
            device_self_time['columns'] = [{'type': 'string', 'name': 'name'}, {'type': 'number', 'name': 'value'}]
            device_self_time['rows'] = op_device_self_time
        else:
            device_self_time = None

        if len(op_host_total_time) > 0:
            host_total_time['title'] = 'Host Total Time (us)'
            host_total_time['columns'] = [{'type': 'string', 'name': 'name'}, {'type': 'number', 'name': 'value'}]
            host_total_time['rows'] = op_host_total_time
        else:
            host_total_time = None

        if len(op_host_self_time) > 0:
            host_self_time['title'] = 'Host Self Time (us)'
            host_self_time['columns'] = [{'type': 'string', 'name': 'name'}, {'type': 'number', 'name': 'value'}]
            host_self_time['rows'] = op_host_self_time
        else:
            host_self_time = None

        data['device_total_time'] = device_total_time
        data['device_self_time'] = device_self_time
        data['host_total_time'] = host_total_time
        data['host_self_time'] = host_self_time

        return data

    def _generate_op_table(self, op_list: Iterable[OperatorAgg], group_by_input_shape=False, call_stack=False):
        show_gpu = self.profile_data.has_kernel or self.profile_data.has_memcpy_or_memset

        if group_by_input_shape:
            stack_list_dict = self.profile_data.stack_lists_group_by_name_input
        else:
            stack_list_dict = self.profile_data.stack_lists_group_by_name

        op_list = sorted(op_list,
                         key=lambda x: x.self_device_duration if show_gpu else x.self_host_duration,
                         reverse=True)

        data = list()
        result = {
            'metadata': {
                'sort': 'device_self_duration' if show_gpu else 'host_self_duration',
                'tooltips': {
                    'tc_eligible': consts.TOOLTIP_OP_TC_ELIGIBLE,
                    'tc_self_ratio': consts.TOOLTIP_OP_TC_SELF,
                    'tc_total_ratio': consts.TOOLTIP_OP_TC_TOTAL
                }
            },
            'data': data
        }
        for op in op_list:
            # Whether device_duration & self_device_duration are accurate or not depends on the input tracing data.
            row = dict()
            row['name'] = op.name
            if group_by_input_shape:
                row['input_shape'] = op.input_shape
            row['calls'] = op.calls
            if show_gpu:
                row['device_self_duration'] = round(op.self_device_duration)
                row['device_total_duration'] = round(op.device_duration)
            row['host_self_duration'] = round(op.self_host_duration)
            row['host_total_duration'] = round(op.host_duration)
            row['tc_eligible'] = 'Yes' if op.tc_eligible else 'No'
            row['tc_self_ratio'] = round(100 * op.tc_self_ratio, 2)
            row['tc_total_ratio'] = round(100 * op.tc_total_ratio, 2)
            if call_stack:
                row['call_stack'] = op.callstacks.pop()
            else:
                if group_by_input_shape:
                    key = op.name + '###' + str(op.input_shape)
                else:
                    key = op.name
                row['has_call_stack'] = key in stack_list_dict
            data.append(row)

        return result

    def _generate_op_table_for_stack(self, group_by_input_shape: bool):
        if group_by_input_shape:
            stack_list_dict = self.profile_data.stack_lists_group_by_name_input
        else:
            stack_list_dict = self.profile_data.stack_lists_group_by_name

        result = dict()
        for k, v in stack_list_dict.items():
            result[k] = self._generate_op_table(v, group_by_input_shape, True)
        return result

    def _generate_kernel_op_table(self):
        table = {}
        result = {
            'metadata': {
                'sort': 'Total Duration (us)'
            },
            'data': table
        }
        table['columns'] = [{'type': 'string', 'name': 'Name'},
                            {'type': 'string', 'name': 'Operator'},
                            {'type': 'string', 'name': 'Grid'},
                            {'type': 'string', 'name': 'Block'},
                            {'type': 'number', 'name': 'Register Per Thread'},
                            {'type': 'number', 'name': 'Shared Memory'},
                            {'type': 'string', 'name': 'Kernel Uses Tensor Cores',
                             'tooltip': consts.TOOLTIP_KERNEL_USES_TC},
                            {'type': 'string', 'name': 'Op is Tensor Cores eligible',
                             'tooltip': consts.TOOLTIP_KERNEL_OP_TC_ELIGIBLE}]
        col_names = ['Calls', 'Total Duration (us)', 'Mean Duration (us)', 'Max Duration (us)', 'Min Duration (us)']
        for column in col_names:
            table['columns'].append({'type': 'number', 'name': column})
        gpu_metrics_columns = self.profile_data.gpu_metrics_parser.get_gpu_metrics_columns()
        table['columns'].extend(gpu_metrics_columns)

        table['rows'] = []
        kernel_list: List[KernelAggByNameOp] = sorted(
            self.profile_data.kernel_list_groupby_name_op, key=lambda x: x.total_duration, reverse=True)
        for agg_by_name_op in kernel_list:
            kernel_op_row = [agg_by_name_op.name, agg_by_name_op.op_name,
                             str(agg_by_name_op.grid), str(agg_by_name_op.block),
                             str(agg_by_name_op.regs_per_thread or '0'), str(agg_by_name_op.shared_memory or '0'),
                             'Yes' if agg_by_name_op.tc_used else 'No',
                             'Yes' if agg_by_name_op.op_tc_eligible else 'No',
                             agg_by_name_op.calls,
                             agg_by_name_op.total_duration, round(agg_by_name_op.avg_duration),
                             agg_by_name_op.max_duration, agg_by_name_op.min_duration]
            if self.profile_data.gpu_metrics_parser.has_blocks_per_sm:
                kernel_op_row.append(round(agg_by_name_op.avg_blocks_per_sm, 2))
            if self.profile_data.gpu_metrics_parser.has_occupancy:
                kernel_op_row.append(round(agg_by_name_op.avg_occupancy, 2))
            table['rows'].append(kernel_op_row)
        return result

    def _generate_kernel_pie(self):
        pie = {'columns': [{'type': 'string', 'name': 'name'}, {'type': 'number', 'name': 'value'}], 'rows': []}
        for _id, (name, row) in enumerate(self.profile_data.kernel_stat.iterrows()):
            pie['rows'].append([name, row['sum']])
        data = {'total': pie}
        return data

    def _generate_kernel_table(self):
        table = {}
        result = {
            'metadata': {
                'sort': 'Total Duration (us)'
            },
            'data': table
        }
        table['columns'] = [{'type': 'string', 'name': 'Name'},
                            {'type': 'string', 'name': 'Tensor Cores Used',
                             'tooltip': consts.TOOLTIP_KERNEL_USES_TC}]
        columns = ['count', 'sum', 'mean', 'max', 'min']
        round_digits = [0, 0, 0, 0, 0]
        if self.profile_data.gpu_metrics_parser.has_blocks_per_sm:
            columns.append('blocks_per_sm')
            round_digits.append(2)
        if self.profile_data.gpu_metrics_parser.has_occupancy:
            columns.append('occupancy')
            round_digits.append(2)
        col_names = ['Calls', 'Total Duration (us)', 'Mean Duration (us)', 'Max Duration (us)', 'Min Duration (us)']
        for column in col_names:
            table['columns'].append({'type': 'number', 'name': column})
        gpu_metrics_columns = self.profile_data.gpu_metrics_parser.get_gpu_metrics_columns()
        table['columns'].extend(gpu_metrics_columns)

        table['rows'] = []
        for _id, (name, row) in enumerate(self.profile_data.kernel_stat.iterrows()):
            kernel_row = [name, 'Yes' if row['tc_used'] else 'No']
            for i, column in enumerate(columns):
                kernel_row.append(round(row[column]) if round_digits[i] == 0
                                  else round(row[column], round_digits[i]))
            table['rows'].append(kernel_row)
        return result

    def _generate_tc_pie(self):
        pie = {'columns': [{'type': 'string', 'name': 'name'}, {'type': 'number', 'name': 'value'}], 'rows': []}
        pie['rows'].append(['Using Tensor Cores', self.profile_data.tc_used_ratio])
        pie['rows'].append(['Not Using Tensor Cores', 1.0 - self.profile_data.tc_used_ratio])
        data = {'total': pie}
        return data

    @staticmethod
    def _get_gpu_info(device_props, gpu_id):
        if (device_props is None) or (gpu_id >= len(device_props)) or (gpu_id < 0):
            return None

        device_prop: Dict = device_props[gpu_id]
        gpu_info = {}
        name = device_prop.get('name')
        if name is not None:
            gpu_info['Name'] = name

        mem = device_prop.get('totalGlobalMem')
        if mem is not None:
            gpu_info['Memory'] = '{} GB'.format(round(float(mem) / 1024 / 1024 / 1024, 2))
            gpu_info['Memory Raw'] = mem

        major = device_prop.get('computeMajor')
        minor = device_prop.get('computeMinor')
        if major is not None and minor is not None:
            gpu_info['Compute Capability'] = '{}.{}'.format(major, minor)

        return gpu_info


class DistributedRunGenerator(object):
    def __init__(self, all_profile_data: Iterable[DistributedRunProfileData], span):
        self.all_profile_data = all_profile_data
        self.span = span

    def generate_run_profile(self):
        profile_run = DistributedRunProfile(self.span)
        profile_run.views.append(consts.DISTRIBUTED_VIEW)
        profile_run.gpu_info = self._generate_gpu_info()
        profile_run.steps_to_overlap = self._generate_overlap_graph()
        profile_run.steps_to_wait = self._generate_wait_graph()
        profile_run.comm_ops = self._generate_ops_table()
        return profile_run

    def _generate_gpu_info(self):
        # first key is node name, the second key is process id, the third key is GPU0/,
        # the value is the gpu info json
        result: Dict[str, Dict[str, Dict[str, Dict]]] = OrderedDict()
        index = 0
        for data in sorted(self.all_profile_data, key=lambda x: x.worker):
            if not data.device_props:
                continue

            match = consts.NODE_PROCESS_PATTERN.match(data.worker)
            if match:
                node = match.group(1)
                process_id = match.group(2)
            else:
                logger.warning('cannot parse node name from worker name {}'.format(data.worker))
                node = data.worker
                process_id = index
                index += 1
            if node not in result:
                result[node] = OrderedDict()

            process_id = 'Process ' + str(process_id)
            result[node][process_id] = OrderedDict()
            for used_device in data.used_devices:
                gpu_info = RunGenerator._get_gpu_info(data.device_props, used_device)
                if gpu_info is not None:
                    result[node][process_id]['GPU'+str(used_device)] = gpu_info

        if result:
            for k, v in result.items():
                result[k] = OrderedDict(sorted(v.items()))
            return {
                'metadata': {'title': 'Device Information'},
                'data': result
            }
        else:
            return None

    def _generate_overlap_graph(self):
        result = dict()
        result['metadata'] = {
            'title': 'Computation/Communication Overview',
            'legends': ['Computation', 'Overlapping', 'Communication', 'Other'],
            'units': 'us'
        }
        steps_to_overlap: Dict[str, Dict[str, List[int]]] = OrderedDict()
        steps_to_overlap['all'] = OrderedDict()
        for data in self.all_profile_data:
            steps_to_overlap['all'][data.worker] = [0, 0, 0, 0]
            step_number = len(data.steps_names)
            for i, step_name in enumerate(data.steps_names):
                steps_to_overlap.setdefault(step_name, OrderedDict())
                costs = data.comm_overlap_costs[i]
                steps_to_overlap[step_name][data.worker] = [
                    costs.computation - costs.overlap,
                    costs.overlap,
                    costs.communication - costs.overlap,
                    costs.other
                ]
                steps_to_overlap['all'][data.worker] = [
                    sum(x) for x in zip(steps_to_overlap['all'][data.worker], steps_to_overlap[step_name][data.worker])]
            steps_to_overlap['all'][data.worker] = [x/step_number for x in steps_to_overlap['all'][data.worker]]
        for k, v in steps_to_overlap.items():
            steps_to_overlap[k] = OrderedDict(sorted(v.items()))
        result['data'] = steps_to_overlap
        return result

    def _generate_wait_graph(self):
        result = dict()
        result['metadata'] = {
            'title': 'Synchronizing/Communication Overview',
            'legends': ['Data Transfer Time', 'Synchronizing Time'],
            'units': 'us'
        }
        steps_to_wait: Dict[str, Dict[str, List[int]]] = OrderedDict()

        steps_to_wait['all'] = OrderedDict()
        for data in self.all_profile_data:
            steps_to_wait['all'][data.worker] = [0, 0]
            step_number = len(data.step_comm_stats.values())
            for step, comm_stats in data.step_comm_stats.items():
                steps_to_wait.setdefault(step, OrderedDict())[data.worker] = [
                    comm_stats[1],
                    comm_stats[0]-comm_stats[1]
                ]
                steps_to_wait['all'][data.worker] = [
                    sum(x) for x in zip(steps_to_wait['all'][data.worker], steps_to_wait[step][data.worker])]
            steps_to_wait['all'][data.worker] = [x/step_number for x in steps_to_wait['all'][data.worker]]

        for k, v in steps_to_wait.items():
            steps_to_wait[k] = OrderedDict(sorted(v.items()))
        result['data'] = steps_to_wait
        return result

    def _generate_ops_table(self):
        result = dict()
        result['metadata'] = {'title': 'Communication Operations Stats'}
        workers_to_comm_ops = OrderedDict()
        # Ignore the span for distributed view
        for data in self.all_profile_data:
            table = {}
            table['columns'] = [{'type': 'string', 'name': 'Name'}]
            col_names = [
                'Calls',
                'Total Size (bytes)',
                'Avg Size (bytes)',
                'Total Latency (us)',
                'Avg Latency (us)',
                'Data Transfer Time (us)',
                'Avg Data Transfer Time (us)'
            ]
            for column in col_names:
                table['columns'].append({'type': 'number', 'name': column})
            table['rows'] = []
            for op, stats in data.total_comm_stats.items():
                row = [
                    op,
                    stats[0],
                    stats[1],
                    round(stats[1]/stats[0]),
                    stats[2],
                    round(stats[2]/stats[0]),
                    stats[3],
                    round(stats[3]/stats[0])
                ]
                table['rows'].append(row)
            workers_to_comm_ops[data.worker] = table
        result['data'] = OrderedDict(sorted(workers_to_comm_ops.items()))
        return result



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# --------------------------------------------------------------------------
import atexit
import gzip
import json
import os
import shutil
import sys
import tempfile
import threading
import time
from collections import OrderedDict
from queue import Queue

import werkzeug
from tensorboard.plugins import base_plugin
from werkzeug import exceptions, wrappers

from . import consts, io, utils
from .profiler import RunLoader
from .run import DistributedRunProfile, Run, RunProfile

logger = utils.get_logger()


def decorate_headers(func):
    def wrapper(*args, **kwargs):
        headers = func(*args, **kwargs)
        headers.extend(TorchProfilerPlugin.headers)
        return headers
    return wrapper


exceptions.HTTPException.get_headers = decorate_headers(exceptions.HTTPException.get_headers)


class TorchProfilerPlugin(base_plugin.TBPlugin):
    """TensorBoard plugin for Torch Profiler."""

    plugin_name = consts.PLUGIN_NAME
    headers = [('X-Content-Type-Options', 'nosniff')]
    CONTENT_TYPE = 'application/json'

    def __init__(self, context: base_plugin.TBContext):
        """Instantiates TorchProfilerPlugin.
        Args:
          context: A base_plugin.TBContext instance.
        """
        super(TorchProfilerPlugin, self).__init__(context)
        self.logdir = io.abspath(context.logdir.rstrip('/'))

        self._load_lock = threading.Lock()
        self._load_threads = []

        self._runs = OrderedDict()
        self._runs_lock = threading.Lock()

        self._temp_dir = tempfile.mkdtemp()
        self._cache = io.Cache(self._temp_dir)
        self._queue = Queue()
        self._gpu_metrics_file_dict = {}
        monitor_runs = threading.Thread(target=self._monitor_runs, name='monitor_runs', daemon=True)
        monitor_runs.start()

        receive_runs = threading.Thread(target=self._receive_runs, name='receive_runs', daemon=True)
        receive_runs.start()

        self.diff_run_cache = {}
        self.diff_run_flatten_cache = {}

        def clean():
            logger.debug('starting cleanup...')
            self._cache.__exit__(*sys.exc_info())
            logger.debug('remove temporary cache directory %s' % self._temp_dir)
            shutil.rmtree(self._temp_dir)

        atexit.register(clean)

    def is_active(self):
        """Returns whether there is relevant data for the plugin to process.
        If there is no any pending run, hide the plugin
        """
        if self.is_loading:
            return True
        else:
            with self._runs_lock:
                return bool(self._runs)

    def get_plugin_apps(self):
        return {
            '/index.js': self.static_file_route,
            '/index.html': self.static_file_route,
            '/trace_viewer_full.html': self.static_file_route,
            '/trace_embedding.html': self.static_file_route,
            '/runs': self.runs_route,
            '/views': self.views_route,
            '/workers': self.workers_route,
            '/spans': self.spans_route,
            '/overview': self.overview_route,
            '/operation': self.operation_pie_route,
            '/operation/table': self.operation_table_route,
            '/operation/stack': self.operation_stack_route,
            '/kernel': self.kernel_pie_route,
            '/kernel/table': self.kernel_table_route,
            '/kernel/tc_pie': self.kernel_tc_route,
            '/trace': self.trace_route,
            '/distributed/gpuinfo': self.dist_gpu_info_route,
            '/distributed/overlap': self.comm_overlap_route,
            '/distributed/waittime': self.comm_wait_route,
            '/distributed/commops': self.comm_ops_route,
            '/memory': self.memory_route,
            '/memory_curve': self.memory_curve_route,
            '/memory_events': self.memory_events_route,
            '/module': self.module_route,
            '/tree': self.op_tree_route,
            '/diff': self.diff_run_route,
            '/diffnode': self.diff_run_node_route,
        }

    def frontend_metadata(self):
        return base_plugin.FrontendMetadata(es_module_path='/index.js', disable_reload=True)

    @wrappers.Request.application
    def runs_route(self, request: werkzeug.Request):
        with self._runs_lock:
            names = list(self._runs.keys())

        data = {
            'runs': names,
            'loading': self.is_loading
        }
        return self.respond_as_json(data)

    @wrappers.Request.application
    def views_route(self, request: werkzeug.Request):
        name = request.args.get('run')
        self._validate(run=name)
        run = self._get_run(name)
        views_list = [view.display_name for view in run.views]
        return self.respond_as_json(views_list)

    @wrappers.Request.application
    def workers_route(self, request: werkzeug.Request):
        name = request.args.get('run')
        view = request.args.get('view')
        self._validate(run=name, view=view)
        run = self._get_run(name)
        return self.respond_as_json(run.get_workers(view))

    @wrappers.Request.application
    def spans_route(self, request: werkzeug.Request):
        name = request.args.get('run')
        worker = request.args.get('worker')
        self._validate(run=name, worker=worker)
        run = self._get_run(name)
        return self.respond_as_json(run.get_spans(worker))

    @wrappers.Request.application
    def overview_route(self, request: werkzeug.Request):
        profile = self._get_profile_for_request(request)
        name = request.args.get('run')
        run = self._get_run(name)
        data = profile.overview
        is_gpu_used = profile.has_runtime or profile.has_kernel or profile.has_memcpy_or_memset
        normal_workers = [worker for worker in run.workers if worker != 'All']
        data['environments'] = [{'title': 'Number of Worker(s)', 'value': str(len(normal_workers))},
                                {'title': 'Device Type', 'value': 'GPU' if is_gpu_used else 'CPU'}]
        if profile.gpu_summary and profile.gpu_tooltip:
            data['gpu_metrics'] = {'title': 'GPU Summary',
                                   'data': profile.gpu_summary,
                                   'tooltip': profile.gpu_tooltip}

        return self.respond_as_json(data)

    @wrappers.Request.application
    def operation_pie_route(self, request: werkzeug.Request):
        profile = self._get_profile_for_request(request)

        group_by = request.args.get('group_by')
        if group_by == 'OperationAndInputShape':
            return self.respond_as_json(profile.operation_pie_by_name_input)
        else:
            return self.respond_as_json(profile.operation_pie_by_name)

    @wrappers.Request.application
    def operation_table_route(self, request: werkzeug.Request):
        profile = self._get_profile_for_request(request)

        group_by = request.args.get('group_by')
        if group_by == 'OperationAndInputShape':
            return self.respond_as_json(profile.operation_table_by_name_input)
        else:
            return self.respond_as_json(profile.operation_table_by_name)

    @wrappers.Request.application
    def operation_stack_route(self, request: werkzeug.Request):
        profile = self._get_profile_for_request(request)

        op_name = request.args.get('op_name')
        self._validate(op_name=op_name)
        group_by = request.args.get('group_by')
        input_shape = request.args.get('input_shape')
        if group_by == 'OperationAndInputShape':
            return self.respond_as_json(profile.operation_stack_by_name_input[str(op_name)+'###'+str(input_shape)])
        else:
            return self.respond_as_json(profile.operation_stack_by_name[str(op_name)])

    @wrappers.Request.application
    def kernel_pie_route(self, request: werkzeug.Request):
        profile = self._get_profile_for_request(request)

        return self.respond_as_json(profile.kernel_pie)

    @wrappers.Request.application
    def kernel_table_route(self, request: werkzeug.Request):
        profile = self._get_profile_for_request(request)

        group_by = request.args.get('group_by')
        if group_by == 'Kernel':
            return self.respond_as_json(profile.kernel_table)
        else:
            return self.respond_as_json(profile.kernel_op_table)

    @wrappers.Request.application
    def kernel_tc_route(self, request: werkzeug.Request):
        profile = self._get_profile_for_request(request)

        return self.respond_as_json(profile.tc_pie)

    @wrappers.Request.application
    def trace_route(self, request: werkzeug.Request):
        profile = self._get_profile_for_request(request)

        if not profile.has_kernel:  # Pure CPU.
            raw_data = self._cache.read(profile.trace_file_path)
            if not profile.trace_file_path.endswith('.gz'):
                raw_data = gzip.compress(raw_data, 1)
        else:
            file_with_gpu_metrics = self._gpu_metrics_file_dict.get(profile.trace_file_path)
            if file_with_gpu_metrics:
                raw_data = io.read(file_with_gpu_metrics)
            else:
                raw_data = self._cache.read(profile.trace_file_path)
                if profile.trace_file_path.endswith('.gz'):
                    raw_data = gzip.decompress(raw_data)
                raw_data = profile.append_gpu_metrics(raw_data)

                # write the data to temp file
                fp = tempfile.NamedTemporaryFile('w+b', suffix='.json.gz', dir=self._temp_dir, delete=False)
                fp.close()
                # Already compressed, no need to gzip.open
                with open(fp.name, mode='wb') as file:
                    file.write(raw_data)
                self._gpu_metrics_file_dict[profile.trace_file_path] = fp.name

        headers = [('Content-Encoding', 'gzip')]
        headers.extend(TorchProfilerPlugin.headers)
        return werkzeug.Response(raw_data, content_type=TorchProfilerPlugin.CONTENT_TYPE, headers=headers)

    @wrappers.Request.application
    def dist_gpu_info_route(self, request: werkzeug.Request):
        profile = self._get_distributed_profile_for_request(request)
        return self.respond_as_json(profile.gpu_info)

    @wrappers.Request.application
    def comm_overlap_route(self, request: werkzeug.Request):
        profile = self._get_distributed_profile_for_request(request)
        return self.respond_as_json(profile.steps_to_overlap)

    @wrappers.Request.application
    def comm_wait_route(self, request: werkzeug.Request):
        profile = self._get_distributed_profile_for_request(request)
        return self.respond_as_json(profile.steps_to_wait)

    @wrappers.Request.application
    def comm_ops_route(self, request: werkzeug.Request):
        profile = self._get_distributed_profile_for_request(request)
        return self.respond_as_json(profile.comm_ops)

    @wrappers.Request.application
    def memory_route(self, request: werkzeug.Request):
        profile = self._get_profile_for_request(request)
        start_ts = request.args.get('start_ts', None)
        end_ts = request.args.get('end_ts', None)
        memory_metric = request.args.get('memory_metric', 'KB')
        if start_ts is not None:
            start_ts = int(start_ts)
        if end_ts is not None:
            end_ts = int(end_ts)

        return self.respond_as_json(
            profile.get_memory_stats(start_ts=start_ts, end_ts=end_ts, memory_metric=memory_metric), True)

    @wrappers.Request.application
    def memory_curve_route(self, request: werkzeug.Request):
        profile = self._get_profile_for_request(request)
        time_metric = request.args.get('time_metric', 'ms')
        memory_metric = request.args.get('memory_metric', 'MB')
        return self.respond_as_json(
            profile.get_memory_curve(time_metric=time_metric, memory_metric=memory_metric), True)

    @wrappers.Request.application
    def memory_events_route(self, request: werkzeug.Request):
        profile = self._get_profile_for_request(request)
        start_ts = request.args.get('start_ts', None)
        end_ts = request.args.get('end_ts', None)
        time_metric = request.args.get('time_metric', 'ms')
        memory_metric = request.args.get('memory_metric', 'KB')
        if start_ts is not None:
            start_ts = int(start_ts)
        if end_ts is not None:
            end_ts = int(end_ts)

        return self.respond_as_json(
            profile.get_memory_events(start_ts, end_ts, time_metric=time_metric,
                                      memory_metric=memory_metric), True)

    @wrappers.Request.application
    def module_route(self, request: werkzeug.Request):
        profile = self._get_profile_for_request(request)
        content = profile.get_module_view()
        if content:
            return self.respond_as_json(content, True)
        else:
            name = request.args.get('run')
            worker = request.args.get('worker')
            span = request.args.get('span')
            raise exceptions.NotFound('could not find the run for %s/%s/%s' % (name, worker, span))

    @wrappers.Request.application
    def op_tree_route(self, request: werkzeug.Request):
        profile = self._get_profile_for_request(request)
        content = profile.get_operator_tree()
        return self.respond_as_json(content, True)

    @wrappers.Request.application
    def diff_run_route(self, request: werkzeug.Request):
        base, exp = self.get_diff_runs(request)
        diff_stats = self.get_diff_status(base, exp)
        content = diff_stats.get_diff_tree_summary()
        return self.respond_as_json(content, True)

    @wrappers.Request.application
    def diff_run_node_route(self, request: werkzeug.Request):
        base, exp = self.get_diff_runs(request)
        path = request.args.get('path', '0')
        stats_dict = self.get_diff_stats_dict(base, exp)
        diff_stat = stats_dict.get(path)
        if diff_stat is None:
            raise exceptions.NotFound('could not find diff run for %s' % (path))
        content = diff_stat.get_diff_node_summary(path)
        return self.respond_as_json(content, True)

    @wrappers.Request.application
    def static_file_route(self, request: werkzeug.Request):
        filename = os.path.basename(request.path)
        extension = os.path.splitext(filename)[1]
        if extension == '.html':
            mimetype = 'text/html'
        elif extension == '.css':
            mimetype = 'text/css'
        elif extension == '.js':
            mimetype = 'application/javascript'
        else:
            mimetype = 'application/octet-stream'
        filepath = os.path.join(os.path.dirname(__file__), 'static', filename)
        try:
            with open(filepath, 'rb') as infile:
                contents = infile.read()
        except IOError:
            raise exceptions.NotFound('404 Not Found')
        return werkzeug.Response(
            contents, content_type=mimetype, headers=TorchProfilerPlugin.headers
        )

    @staticmethod
    def respond_as_json(obj, compress: bool = False):
        content = json.dumps(obj)
        headers = []
        headers.extend(TorchProfilerPlugin.headers)
        if compress:
            content_bytes = content.encode('utf-8')
            raw_data = gzip.compress(content_bytes, 1)
            headers.append(('Content-Encoding', 'gzip'))
            return werkzeug.Response(raw_data, content_type=TorchProfilerPlugin.CONTENT_TYPE, headers=headers)
        else:
            return werkzeug.Response(content, content_type=TorchProfilerPlugin.CONTENT_TYPE, headers=headers)

    @property
    def is_loading(self):
        with self._load_lock:
            return bool(self._load_threads)

    def get_diff_runs(self, request: werkzeug.Request):
        name = request.args.get('run')
        span = request.args.get('span')
        worker = request.args.get('worker')
        self._validate(run=name, worker=worker, span=span)
        base = self._get_profile(name, worker, span)

        exp_name = request.args.get('exp_run')
        exp_span = request.args.get('exp_span')
        exp_worker = request.args.get('exp_worker')
        self._validate(exp_run=exp_name, exp_worker=exp_worker, exp_span=exp_span)
        exp = self._get_profile(exp_name, exp_worker, exp_span)

        return base, exp

    def get_diff_status(self, base: RunProfile, exp: RunProfile):
        key = (base, exp)
        diff_stats = self.diff_run_cache.get(key)
        if diff_stats is None:
            diff_stats = base.compare_run(exp)
            self.diff_run_cache[key] = diff_stats

        return diff_stats

    def get_diff_stats_dict(self, base: RunProfile, exp: RunProfile):
        key = (base, exp)
        stats_dict = self.diff_run_flatten_cache.get(key)
        if stats_dict is None:
            diff_stats = self.get_diff_status(base, exp)
            stats_dict = diff_stats.flatten_diff_tree()
            self.diff_run_flatten_cache[key] = stats_dict
        return stats_dict

    def _monitor_runs(self):
        logger.info('Monitor runs begin')

        try:
            touched = set()
            while True:
                try:
                    logger.debug('Scan run dir')
                    run_dirs = self._get_run_dirs()

                    has_dir = False
                    # Assume no deletion on run directories, trigger async load if find a new run
                    for run_dir in run_dirs:
                        has_dir = True
                        if run_dir not in touched:
                            touched.add(run_dir)
                            logger.info('Find run directory %s', run_dir)
                            # Use threading to avoid UI stall and reduce data parsing time
                            t = threading.Thread(target=self._load_run, args=(run_dir,))
                            t.start()
                            with self._load_lock:
                                self._load_threads.append(t)

                    if not has_dir:
                        # handle directory removed case.
                        self._runs.clear()
                except Exception as ex:
                    logger.warning('Failed to scan runs. Exception=%s', ex, exc_info=True)

                time.sleep(consts.MONITOR_RUN_REFRESH_INTERNAL_IN_SECONDS)
        except Exception:
            logger.exception('Failed to start monitor_runs')

    def _receive_runs(self):
        while True:
            run: Run = self._queue.get()
            if run is None:
                continue

            logger.info('Add run %s', run.name)
            with self._runs_lock:
                is_new = run.name not in self._runs
                self._runs[run.name] = run
                if is_new:
                    self._runs = OrderedDict(sorted(self._runs.items()))

    def _get_run_dirs(self):
        """Scan logdir, find PyTorch Profiler run directories.
        A directory is considered to be a run if it contains 1 or more *.pt.trace.json[.gz].
        E.g. there are 2 runs: run1, run2
            /run1
                /[worker1].pt.trace.json.gz
                /[worker2].pt.trace.json.gz
            /run2
                /[worker1].pt.trace.json
        """
        for root, _, files in io.walk(self.logdir):
            for file in files:
                if utils.is_chrome_trace_file(file):
                    yield root
                    break

    def _load_run(self, run_dir):
        try:
            name = self._get_run_name(run_dir)
            logger.info('Load run %s', name)
            # Currently, assume run data is immutable, so just load once
            loader = RunLoader(name, run_dir, self._cache)
            run = loader.load()
            logger.info('Run %s loaded', name)
            self._queue.put(run)
        except Exception as ex:
            logger.warning('Failed to load run %s. Exception=%s', ex, name, exc_info=True)

        t = threading.current_thread()
        with self._load_lock:
            try:
                self._load_threads.remove(t)
            except ValueError:
                logger.warning('could not find the thread {}'.format(run_dir))

    def _get_run(self, name) -> Run:
        with self._runs_lock:
            run = self._runs.get(name, None)

        if run is None:
            raise exceptions.NotFound('could not find the run for %s' % (name))

        return run

    def _get_run_name(self, run_dir):
        logdir = io.abspath(self.logdir)
        if run_dir == logdir:
            name = io.basename(run_dir)
        else:
            name = io.relpath(run_dir, logdir)
        return name

    def _get_profile_for_request(self, request: werkzeug.Request) -> RunProfile:
        name = request.args.get('run')
        span = request.args.get('span')
        worker = request.args.get('worker')
        self._validate(run=name, worker=worker)
        profile = self._get_profile(name, worker, span)
        if not isinstance(profile, RunProfile):
            raise exceptions.BadRequest('Get an unexpected profile type %s for %s/%s' % (type(profile), name, worker))

        return profile

    def _get_distributed_profile_for_request(self, request: werkzeug.Request) -> DistributedRunProfile:
        name = request.args.get('run')
        span = request.args.get('span')
        self._validate(run=name)
        profile = self._get_profile(name, 'All', span)
        if not isinstance(profile, DistributedRunProfile):
            raise exceptions.BadRequest('Get an unexpected distributed profile type %s for %s' % (type(profile), name))

        return profile

    def _get_profile(self, name, worker, span):
        run = self._get_run(name)
        profile = run.get_profile(worker, span)
        if profile is None:
            raise exceptions.NotFound('could not find the profile for %s/%s/%s ' % (name, worker, span))
        return profile

    def _validate(self, **kwargs):
        for name, v in kwargs.items():
            if v is None:
                raise exceptions.BadRequest('Must specify %s in request url' % (name))



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# -------------------------------------------------------------------------
import sys
from typing import Generator, List, Union

from ..node import (BackwardNode, DataLoaderNode, ModuleNode, OperatorNode,
                    OptimizerNode, ProfilerStepNode)
from .contract import DiffStats, OpStats
from .operator import Operator, Operators, create_operator

INDENT = '    '
RUN_NODE_TYPES = (BackwardNode, DataLoaderNode, ModuleNode, OptimizerNode, ProfilerStepNode)


class DiffNode:
    def __init__(self, left: Operator, right: Operator):
        self.left: Operator = left
        self.right: Operator = right
        self.children: List[DiffNode] = []

    def build_tree(self):
        '''build the children from the left_node and right_node'''
        if not isinstance(self.left, Operators) or not isinstance(self.right, Operators):
            # TODO: do we need calculate the stats or not?
            return

        if isinstance(self.left.op_nodes, OperatorNode) and isinstance(self.right.op_nodes, OperatorNode):
            # simple node match.
            diff_nodes = list(DiffNode.compare_operator_nodes(
                self.left.op_nodes.children, self.right.op_nodes.children))
            if diff_nodes:
                self.children.extend(diff_nodes)
        elif isinstance(self.left.op_nodes, list) and isinstance(self.right.op_nodes, list):
            # compare two list
            diff_nodes = list(DiffNode.compare_operator_nodes(self.left.op_nodes, self.right.op_nodes))
            if diff_nodes:
                self.children.extend(diff_nodes)
        else:
            # one single item and one list
            pass
            # TODO: do we need statistic the stats for both operator and kernel here?

    @staticmethod
    def create_node(
            left: Union[OperatorNode, List[OperatorNode]],
            right: Union[OperatorNode, List[OperatorNode]]) -> 'DiffNode':
        if isinstance(left, list) and len(left) == 1:
            left = left[0]
        if isinstance(right, list) and len(right) == 1:
            right = right[0]

        node = DiffNode(create_operator(left), create_operator(right))
        node.build_tree()
        return node

    @staticmethod
    def compare_operator_nodes(
            left_nodes: List[OperatorNode],
            right_nodes: List[OperatorNode]) -> Generator['DiffNode', None, None]:
        '''Given two OperatorNode lists, find the DataLoader/Module/Backward/Optimizer node and create the child list DiffNode
        '''
        right_keys = [(type(r), r.name) for r in right_nodes]

        # find matching points in the two list
        matched_paris = []
        key_index = 0
        for i, left_node in enumerate(left_nodes):
            if not isinstance(left_node, RUN_NODE_TYPES):
                # only handle DataLoader/Module/Backward/Optimizer nodes
                continue

            for j in range(key_index, len(right_keys)):
                if right_keys[j] == (type(left_node), left_node.name):
                    matched_paris.append((i, j))
                    key_index = j + 1
                    break

        if not matched_paris:
            # there is not any matching points.
            return

        # split the two list by the matching points
        l_iter = 0
        r_iter = 0

        for (l, r) in matched_paris:
            left_child = left_nodes[l_iter:l]
            right_child = right_nodes[r_iter:r]
            if left_child or right_child:
                yield DiffNode.create_node(left_child, right_child)

            yield DiffNode.create_node(left_nodes[l], right_nodes[r])
            l_iter = l + 1
            r_iter = r + 1
            # TODO: fill unknown nodes in case of the start_time of next node and current
            # end time is bigger than threshold.
            # Or do we need move the logic into frondend for visualization?

        # process the remaining nodes
        left_remaining = left_nodes[l_iter:]
        right_remaining = right_nodes[r_iter:]
        if left_remaining or right_remaining:
            yield DiffNode.create_node(left_remaining, right_remaining)


def compare_op_tree(left: OperatorNode, right: OperatorNode) -> DiffNode:
    '''Create the diff tree from two root node
       TODO: need handle the different threads case
             need add runtimes besides of children?
    '''
    left_children = list(get_tree_operators(left))
    right_children = list(get_tree_operators(right))
    return DiffNode.create_node(left_children, right_children)


def get_tree_operators(root: OperatorNode) -> Generator[OperatorNode, None, None]:
    '''Get the operators by the root operators by excluding the ProfilerStepNode
    '''
    profiler_nodes = [c for c in root.children if isinstance(c, ProfilerStepNode)]
    if not profiler_nodes:
        # there is no ProfilerStepNode at all
        yield from root.children
    else:
        yield from (child for p in profiler_nodes for child in p.children)


def diff_summary(node: DiffNode) -> DiffStats:
    if not node:
        return None

    left = OpStats(
        node.left.name,
        node.left.duration,
        node.left.device_duration,
        node.left.total_duration,
        list(node.left.aggregate_ops()))
    right = OpStats(
        node.right.name,
        node.right.duration,
        node.right.device_duration,
        node.right.total_duration,
        list(node.right.aggregate_ops()))

    stats = DiffStats(left, right)
    for child in node.children:
        stats.children.append(diff_summary(child))

    return stats


def print_node(node: Union[DiffNode, DiffStats], level: int, index: int, file=sys.stdout):
    file.write(f'{INDENT * level}level {level}, index {index}:\n')
    file.write(f'{INDENT * (level + 1)}left : {node.left}\n')
    file.write(f'{INDENT * (level + 1)}right: {node.right}\n')
    for i, child in enumerate(node.children):
        print_node(child, level + 1, i, file=file)


def print_ops(op: Operators, prefix: str = INDENT, file=sys.stdout):
    if isinstance(op.op_nodes, list):
        for n in op.op_nodes:
            file.write(f'{prefix}{n.name}\n')
    else:
        file.write(f'{prefix}{op.op_nodes.name}\n')



"""
This file is forked from
https://github.com/tensorflow/tensorboard/blob/master/tensorboard/compat/tensorflow_stub/io/gfile.py.
The following functionalities are added after forking:
* Check Azure Blob & Google Cloud available or not
* get_filesystem changes to support Azure Blobs
* add BaseFileSystem and PathBase abstracted class for the filesystem.
* add download_file for each file system to cache the remote file to local temporary folder.
* add AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY for S3 file system which is not supported by tensorboard.
* add Azure blob file system
* add Google Cloud file system
* add specialized walk for Local file system, Azure Blob and Google Cloud to improve the walk performance.
* add global wrapper for abspath, basename, join, download_file.
* change the global walk wrapper to support specialized walk.
"""
import glob as py_glob
import os
import tempfile

from .. import utils
from .base import BaseFileSystem, LocalPath, RemotePath, StatData
from .utils import as_bytes, as_text, parse_blob_url

logger = utils.get_logger()

try:
    import boto3
    import botocore.exceptions

    S3_ENABLED = True
except ImportError:
    S3_ENABLED = False

try:
    from azure.storage.blob import ContainerClient
    BLOB_ENABLED = True
except ImportError:
    BLOB_ENABLED = False

try:
    # Imports the Google Cloud client library
    from google.cloud import storage
    GS_ENABLED = True
except ImportError:
    GS_ENABLED = False


_DEFAULT_BLOCK_SIZE = 16 * 1024 * 1024

# Registry of filesystems by prefix.
#
# Currently supports:
#  * "s3://" URLs for S3 based on boto3
#  * "https://<account>.blob.core.windows.net" for Azure Blob based on azure-storage-blob
#  * "gs://" URLs for Google Cloud based on google-cloud-storage
#  * Local filesystem when not match any prefix.
_REGISTERED_FILESYSTEMS = {}


def register_filesystem(prefix, filesystem):
    if ":" in prefix:
        raise ValueError("Filesystem prefix cannot contain a :")
    _REGISTERED_FILESYSTEMS[prefix] = filesystem


def get_filesystem(filename):
    """Return the registered filesystem for the given file."""
    prefix = ""
    index = filename.find("://")
    if index >= 0:
        prefix = filename[:index]
    if prefix.upper() in ('HTTP', 'HTTPS'):
        root, _ = parse_blob_url(filename)
        if root.lower().endswith('.blob.core.windows.net'):
            fs = _REGISTERED_FILESYSTEMS.get('blob', None)
        else:
            raise ValueError("Not supported file system for prefix %s" % root)
    else:
        fs = _REGISTERED_FILESYSTEMS.get(prefix, None)
    if fs is None:
        raise ValueError("No recognized filesystem for prefix %s" % prefix)
    return fs


class LocalFileSystem(LocalPath, BaseFileSystem):
    def __init__(self):
        pass

    def exists(self, filename):
        return os.path.exists(filename)

    def read(self, filename, binary_mode=False, size=None, continue_from=None):
        mode = "rb" if binary_mode else "r"
        encoding = None if binary_mode else "utf8"
        if not self.exists(filename):
            raise FileNotFoundError(filename)

        offset = None
        if continue_from is not None:
            offset = continue_from.get("opaque_offset", None)
        with open(filename, mode, encoding=encoding) as f:
            if offset is not None:
                f.seek(offset)
            data = f.read(size)
            # The new offset may not be `offset + len(data)`, due to decoding
            # and newline translation.
            # So, just measure it in whatever terms the underlying stream uses.
            continuation_token = {"opaque_offset": f.tell()}
            return (data, continuation_token)

    def write(self, filename, file_content, binary_mode=False):
        """Writes string file contents to a file, overwriting any existing contents.
        """
        self._write(filename, file_content, "wb" if binary_mode else "w")

    def support_append(self):
        return True

    def append(self, filename, file_content, binary_mode=False):
        """Append string file contents to a file.
        """
        self._write(filename, file_content, "ab" if binary_mode else "a")

    def _write(self, filename, file_content, mode):
        encoding = None if "b" in mode else "utf8"
        with open(filename, mode, encoding=encoding) as f:
            compatify = as_bytes if "b" in mode else as_text
            f.write(compatify(file_content))

    def glob(self, filename):
        """Returns a list of files that match the given pattern(s)."""
        if isinstance(filename, str):
            return [
                matching_filename
                for matching_filename in py_glob.glob(filename)
            ]
        else:
            return [
                matching_filename
                for single_filename in filename
                for matching_filename in py_glob.glob(single_filename)
            ]

    def isdir(self, dirname):
        return os.path.isdir(dirname)

    def listdir(self, dirname):
        entries = os.listdir(dirname)
        entries = [item for item in entries]
        return entries

    def makedirs(self, path):
        os.makedirs(path, exist_ok=True)

    def stat(self, filename):
        """Returns file statistics for a given path."""
        # NOTE: Size of the file is given by .st_size as returned from
        # os.stat(), but we convert to .length
        file_length = os.stat(filename).st_size
        return StatData(file_length)

    def walk(self, top, topdown=True, onerror=None):
        # Note on followlinks=True: per the tensorboard documentation [1], users are encouraged to
        # use symlink trees to have fine-grained control over the filesystem layout of runs. To
        # support such trees, we must follow links.
        # [1] https://github.com/tensorflow/tensorboard/blob/master/README.md#logdir--logdir_spec-legacy-mode
        yield from os.walk(top, topdown, onerror, followlinks=True)


class S3FileSystem(RemotePath, BaseFileSystem):
    """Provides filesystem access to S3."""

    def __init__(self):
        if not boto3:
            raise ImportError("boto3 must be installed for S3 support.")
        self._s3_endpoint = os.environ.get("S3_ENDPOINT", None)
        access_key = os.environ.get("AWS_ACCESS_KEY_ID")
        secret_key = os.environ.get("AWS_SECRET_ACCESS_KEY")
        if access_key and secret_key:
            boto3.setup_default_session(
                aws_access_key_id=access_key, aws_secret_access_key=secret_key)

    def bucket_and_path(self, url):
        """Split an S3-prefixed URL into bucket and path."""
        if url.startswith("s3://"):
            url = url[len("s3://"):]
        idx = url.index("/")
        bucket = url[:idx]
        path = url[(idx + 1):]
        return bucket, path

    def exists(self, filename):
        """Determines whether a path exists or not."""
        client = boto3.client("s3", endpoint_url=self._s3_endpoint)
        bucket, path = self.bucket_and_path(filename)
        r = client.list_objects(Bucket=bucket, Prefix=path, Delimiter="/")
        if r.get("Contents") or r.get("CommonPrefixes"):
            return True
        return False

    def read(self, filename, binary_mode=False, size=None, continue_from=None):
        """Reads contents of a file to a string."""
        s3 = boto3.resource("s3", endpoint_url=self._s3_endpoint)
        bucket, path = self.bucket_and_path(filename)
        args = {}

        # S3 use continuation tokens of the form: {byte_offset: number}
        offset = 0
        if continue_from is not None:
            offset = continue_from.get("byte_offset", 0)

        endpoint = ""
        if size is not None:
            endpoint = offset + size

        if offset != 0 or endpoint != "":
            args["Range"] = "bytes={}-{}".format(offset, endpoint)

        logger.info("s3: starting reading file %s" % filename)
        try:
            stream = s3.Object(bucket, path).get(**args)["Body"].read()
        except botocore.exceptions.ClientError as exc:
            if exc.response["Error"]["Code"] in ["416", "InvalidRange"]:
                if size is not None:
                    # Asked for too much, so request just to the end. Do this
                    # in a second request so we don't check length in all cases.
                    client = boto3.client("s3", endpoint_url=self._s3_endpoint)
                    obj = client.head_object(Bucket=bucket, Key=path)
                    content_length = obj["ContentLength"]
                    endpoint = min(content_length, offset + size)
                if offset == endpoint:
                    # Asked for no bytes, so just return empty
                    stream = b""
                else:
                    args["Range"] = "bytes={}-{}".format(offset, endpoint)
                    stream = s3.Object(bucket, path).get(**args)["Body"].read()
            else:
                raise

        logger.info("s3: file %s download is done, size is %d" %
                    (filename, len(stream)))
        # `stream` should contain raw bytes here (i.e., there has been neither decoding nor newline translation),
        # so the byte offset increases by the expected amount.
        continuation_token = {"byte_offset": (offset + len(stream))}
        if binary_mode:
            return (bytes(stream), continuation_token)
        else:
            return (stream.decode("utf-8"), continuation_token)

    def write(self, filename, file_content, binary_mode=False):
        """Writes string file contents to a file."""
        client = boto3.client("s3", endpoint_url=self._s3_endpoint)
        bucket, path = self.bucket_and_path(filename)
        if binary_mode:
            if not isinstance(file_content, bytes):
                raise TypeError("File content type must be bytes")
        else:
            file_content = as_bytes(file_content)
        client.put_object(Body=file_content, Bucket=bucket, Key=path)

    def download_file(self, file_to_download, file_to_save):
        logger.info("s3: starting downloading file %s as %s" %
                    (file_to_download, file_to_save))
        # Use boto3.resource instead of boto3.client('s3') to support minio.
        # https://docs.min.io/docs/how-to-use-aws-sdk-for-python-with-minio-server.html
        # To support minio, the S3_ENDPOINT need to be set like: S3_ENDPOINT=http://localhost:9000
        s3 = boto3.resource("s3", endpoint_url=self._s3_endpoint)
        bucket, path = self.bucket_and_path(file_to_download)
        s3.Bucket(bucket).download_file(path, file_to_save)
        logger.info("s3: file %s is downloaded as %s" % (file_to_download, file_to_save))
        return

    def glob(self, filename):
        """Returns a list of files that match the given pattern(s)."""
        # Only support prefix with * at the end and no ? in the string
        star_i = filename.find("*")
        quest_i = filename.find("?")
        if quest_i >= 0:
            raise NotImplementedError("{} not supported".format(filename))
        if star_i != len(filename) - 1:
            return []

        filename = filename[:-1]
        client = boto3.client("s3", endpoint_url=self._s3_endpoint)
        bucket, path = self.bucket_and_path(filename)
        p = client.get_paginator("list_objects")
        keys = []
        for r in p.paginate(Bucket=bucket, Prefix=path):
            for o in r.get("Contents", []):
                key = o["Key"][len(path):]
                if key:
                    keys.append(filename + key)
        return keys

    def isdir(self, dirname):
        """Returns whether the path is a directory or not."""
        client = boto3.client("s3", endpoint_url=self._s3_endpoint)
        bucket, path = self.bucket_and_path(dirname)
        if not path.endswith("/"):
            path += "/"
        r = client.list_objects(Bucket=bucket, Prefix=path, Delimiter="/")
        if r.get("Contents") or r.get("CommonPrefixes"):
            return True
        return False

    def listdir(self, dirname):
        """Returns a list of entries contained within a directory."""
        client = boto3.client("s3", endpoint_url=self._s3_endpoint)
        bucket, path = self.bucket_and_path(dirname)
        p = client.get_paginator("list_objects")
        if not path.endswith("/"):
            path += "/"
        keys = []
        for r in p.paginate(Bucket=bucket, Prefix=path, Delimiter="/"):
            keys.extend(
                o["Prefix"][len(path): -1] for o in r.get("CommonPrefixes", [])
            )
            for o in r.get("Contents", []):
                key = o["Key"][len(path):]
                if key:
                    keys.append(key)
        return keys

    def makedirs(self, dirname):
        """Creates a directory and all parent/intermediate directories."""
        if not self.exists(dirname):
            client = boto3.client("s3", endpoint_url=self._s3_endpoint)
            bucket, path = self.bucket_and_path(dirname)
            if not path.endswith("/"):
                path += "/"
            client.put_object(Body="", Bucket=bucket, Key=path)

    def stat(self, filename):
        """Returns file statistics for a given path."""
        # Size of the file is given by ContentLength from S3
        client = boto3.client("s3", endpoint_url=self._s3_endpoint)
        bucket, path = self.bucket_and_path(filename)

        obj = client.head_object(Bucket=bucket, Key=path)
        return StatData(obj["ContentLength"])


register_filesystem("", LocalFileSystem())
if S3_ENABLED:
    register_filesystem("s3", S3FileSystem())

if BLOB_ENABLED:
    from .azureblob import AzureBlobSystem
    register_filesystem("blob", AzureBlobSystem())

if GS_ENABLED:
    from .gs import GoogleBlobSystem
    register_filesystem("gs", GoogleBlobSystem())


class File(object):
    def __init__(self, filename, mode):
        if mode not in ("r", "rb", "br", "w", "wb", "bw"):
            raise ValueError("mode {} not supported by File".format(mode))
        self.filename = filename
        self.fs = get_filesystem(self.filename)
        self.fs_supports_append = self.fs.support_append()
        self.buff = None
        self.buff_chunk_size = _DEFAULT_BLOCK_SIZE
        self.buff_offset = 0
        self.continuation_token = None
        self.write_temp = None
        self.write_started = False
        self.binary_mode = "b" in mode
        self.write_mode = "w" in mode
        self.closed = False

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()
        self.buff = None
        self.buff_offset = 0
        self.continuation_token = None

    def __iter__(self):
        return self

    def _read_buffer_to_offset(self, new_buff_offset):
        old_buff_offset = self.buff_offset
        read_size = min(len(self.buff), new_buff_offset) - old_buff_offset
        self.buff_offset += read_size
        return self.buff[old_buff_offset: old_buff_offset + read_size]

    def read(self, n=None):
        """Reads contents of file to a string.

        Args:
            n: int, number of bytes or characters to read, otherwise
                read all the contents of the file

        Returns:
            Subset of the contents of the file as a string or bytes.
        """
        if self.write_mode:
            raise OSError("File not opened in read mode")

        result = None
        if self.buff and len(self.buff) > self.buff_offset:
            # read from local buffer
            if n is not None:
                chunk = self._read_buffer_to_offset(self.buff_offset + n)
                if len(chunk) == n:
                    return chunk
                result = chunk
                n -= len(chunk)
            else:
                # add all local buffer and update offsets
                result = self._read_buffer_to_offset(len(self.buff))

        # read from filesystem
        read_size = max(self.buff_chunk_size, n) if n is not None else None
        (self.buff, self.continuation_token) = self.fs.read(
            self.filename, self.binary_mode, read_size, self.continuation_token)
        self.buff_offset = 0

        # add from filesystem
        if n is not None:
            chunk = self._read_buffer_to_offset(n)
        else:
            # add all local buffer and update offsets
            chunk = self._read_buffer_to_offset(len(self.buff))
        result = result + chunk if result else chunk

        return result

    def write(self, file_content):
        """Writes string file contents to file, clearing contents of the file
        on first write and then appending on subsequent calls.
        """
        if not self.write_mode:
            raise OSError("File not opened in write mode")

        if self.closed:
            raise OSError("File already closed")

        if self.fs_supports_append:
            if not self.write_started:
                # write the first chunk to truncate file if it already exists
                self.fs.write(self.filename, file_content, self.binary_mode)
                self.write_started = True
            else:
                # append the later chunks
                self.fs.append(self.filename, file_content, self.binary_mode)
        else:
            # add to temp file, but wait for flush to write to final filesystem
            if self.write_temp is None:
                mode = "w+b" if self.binary_mode else "w+"
                self.write_temp = tempfile.TemporaryFile(mode)

            compatify = as_bytes if self.binary_mode else as_text
            self.write_temp.write(compatify(file_content))

    def __next__(self):
        line = None
        while True:
            if not self.buff:
                # read one unit into the buffer
                line = self.read(1)
                if line and (line[-1] == "\n" or not self.buff):
                    return line
                if not self.buff:
                    raise StopIteration()
            else:
                index = self.buff.find("\n", self.buff_offset)
                if index != -1:
                    # include line until now plus newline
                    chunk = self.read(index + 1 - self.buff_offset)
                    line = line + chunk if line else chunk
                    return line

                # read one unit past end of buffer
                chunk = self.read(len(self.buff) + 1 - self.buff_offset)
                line = line + chunk if line else chunk
                if line and (line[-1] == "\n" or not self.buff):
                    return line
                if not self.buff:
                    raise StopIteration()

    def next(self):
        return self.__next__()

    def flush(self):
        if self.closed:
            raise OSError("File already closed")

        if not self.fs_supports_append:
            if self.write_temp is not None:
                # read temp file from the beginning
                self.write_temp.flush()
                self.write_temp.seek(0)
                chunk = self.write_temp.read()
                if chunk is not None:
                    # write full contents and keep in temp file
                    self.fs.write(self.filename, chunk, self.binary_mode)
                    self.write_temp.seek(len(chunk))

    def close(self):
        self.flush()
        if self.write_temp is not None:
            self.write_temp.close()
            self.write_temp = None
            self.write_started = False
        self.closed = True


def exists(filename):
    """Determines whether a path exists or not."""
    return get_filesystem(filename).exists(filename)


def abspath(path):
    return get_filesystem(path).abspath(path)


def basename(path):
    return get_filesystem(path).basename(path)


def relpath(path, start):
    return get_filesystem(path).relpath(path, start)


def join(path, *paths):
    return get_filesystem(path).join(path, *paths)


def download_file(file_to_download, file_to_save):
    """Downloads the file, returning a temporary path to the file after finishing."""
    get_filesystem(file_to_download).download_file(file_to_download, file_to_save)


def glob(filename):
    """Returns a list of files that match the given pattern(s)."""
    return get_filesystem(filename).glob(filename)


def is_local(path):
    """Returns whether the path is a local path"""
    return isinstance(get_filesystem(path), LocalFileSystem)


def isdir(dirname):
    """Returns whether the path is a directory or not."""
    return get_filesystem(dirname).isdir(dirname)


def listdir(dirname):
    """Returns a list of entries contained within a directory.

    The list is in arbitrary order. It does not contain the special entries "."
    and "..".
    """
    return get_filesystem(dirname).listdir(dirname)


def makedirs(path):
    """Creates a directory and all parent/intermediate directories."""
    return get_filesystem(path).makedirs(path)


def walk(top, topdown=True, onerror=None):
    """Recursive directory tree generator for directories.

    Args:
      top: string, a Directory name
      topdown: bool, Traverse pre order if True, post order if False.
      onerror: optional handler for errors. Should be a function, it will be
        called with the error as argument. Rethrowing the error aborts the walk.

    Errors that happen while listing directories are ignored.

    Yields:
      Each yield is a 3-tuple:  the pathname of a directory, followed by lists
      of all its subdirectories and leaf files.
      (dirname, [subdirname, subdirname, ...], [filename, filename, ...])
      as strings
    """
    fs = get_filesystem(top)
    if hasattr(fs, "walk"):
        yield from fs.walk(top, topdown, onerror)
    else:
        top = fs.abspath(top)
        listing = fs.listdir(top)

        files = []
        subdirs = []
        for item in listing:
            full_path = fs.join(top, item)
            if fs.isdir(full_path):
                subdirs.append(item)
            else:
                files.append(item)

        here = (top, subdirs, files)

        if topdown:
            yield here

        for subdir in subdirs:
            joined_subdir = fs.join(top, subdir)
            for subitem in walk(joined_subdir, topdown, onerror=onerror):
                yield subitem

        if not topdown:
            yield here


def stat(filename):
    """Returns file statistics for a given path."""
    return get_filesystem(filename).stat(filename)


def read(file):
    with File(file, 'rb') as f:
        return f.read()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# -------------------------------------------------------------------------
import sys
from collections import defaultdict
from typing import Dict, Iterable, List, Optional, Tuple

from .. import utils
from .node import (BackwardNode, DeviceNode, ModuleNode, OperatorNode,
                   ProfilerStepNode, RuntimeNode, is_operator_node)
from .trace import EventTypes

logger = utils.get_logger()


class OpTreeBuilder:
    BACKWARD_ROOT_PREFIX = 'autograd::engine::evaluate_function:'
    BACKWARD_ACCUMULATE_GRAD = 'autograd::engine::evaluate_function: torch::autograd::AccumulateGrad'

    def __init__(self):
        self.main_tid: int = None
        self.tid2tree: Dict[int, OperatorNode] = None

    def build_tree(self,
                   tid2list: Dict[int, List[OperatorNode]],
                   tid2zero_rt_list: Dict[int, List[RuntimeNode]],
                   staled_device_nodes: List[DeviceNode],
                   fwd_bwd_map: Dict[int, int]):
        """Construct the BackwardNode and replace the original backward nodes
        """
        self.tid2tree = self._build_tree(tid2list, tid2zero_rt_list, staled_device_nodes)

        # if could not find any forward/backward association, skip the processing
        if not fwd_bwd_map:
            logger.debug('there is no any forwarwd backward association, skip processing backward correlation.')
            return self.tid2tree

        self._set_main_tid()

        modules, backward_nodes = self._get_modules()
        if not modules or not backward_nodes:
            return self.tid2tree

        _, ts2parent = OpTreeBuilder._get_node_parents(backward_nodes)
        agg_nodes = OpTreeBuilder._group_backward_nodes(backward_nodes)
        fwd_bwd_root = self._get_backward_roots(fwd_bwd_map, ts2parent, agg_nodes)
        if len(agg_nodes) > 0:
            logger.warning('some nodes cannot find forward nodes')

        backward_modules: List[BackwardNode] = []
        for module in modules:
            OpTreeBuilder._build_backward_module(module, None, fwd_bwd_root, backward_modules)
        OpTreeBuilder._insert_backward_modules(self.tid2tree[self.main_tid], backward_modules)
        self.tid2tree = {tid: root for tid, root in self.tid2tree.items() if len(root.children) > 0}

        return self.tid2tree

    def _build_tree(self, tid2list: Dict[int, List[OperatorNode]], tid2zero_rt_list, staled_device_nodes):
        tid2tree = {}

        for tid, op_list in tid2list.items():
            zero_rt_list = tid2zero_rt_list[tid] if tid in tid2zero_rt_list else []
            # Note that when 2 start_time are equal, the one with bigger end_time should be ahead of the other.
            op_list.sort(key=lambda x: (x.start_time, -x.end_time))
            main_tid = any([op.name.startswith('ProfilerStep#') for op in op_list])
            if main_tid:
                # only append the staled device nodes into main thread
                self.main_tid = op_list[0].tid
                root_node = self._build_tree_internal(op_list, zero_rt_list, tid, staled_device_nodes)
            else:
                root_node = self._build_tree_internal(op_list, zero_rt_list, tid, [])
            tid2tree[int(tid)] = root_node

        return tid2tree

    def _set_main_tid(self):
        if self.main_tid is None and self.tid2tree:
            if len(self.tid2tree) == 1:
                self.main_tid = next(iter(self.tid2tree))
            else:
                # there are multiple tids
                backward_tid = self._find_backward_tid()
                tid2len = {
                    tid: root.end_time - root.start_time for tid, root in self.tid2tree.items()
                    if tid != backward_tid or backward_tid is None
                }
                # get the maximum length as the main thread
                self.main_tid = max(tid2len, key=tid2len.get)

    def _find_backward_tid(self):
        for root in self.tid2tree.values():
            for child in root.children:
                if child.name.startswith(OpTreeBuilder.BACKWARD_ROOT_PREFIX):
                    return child.tid

        return None

    def _build_tree_internal(self, host_node_list, zero_rt_list, tid, staled_device_nodes):
        """host_node_list: list of OperatorNode and ProfilerStepNode.
        zero_rt_list: list of RuntimeNode with external_id=0."""

        def build_tree_relationship(host_node_list: Iterable[OperatorNode], zero_rt_list, staled_device_nodes):
            dummpy_rt: List[RuntimeNode] = []
            if staled_device_nodes:
                # Note: Although kernels of this dummy runtime is put under main thread's tree,
                # we don't know which thread launches them.
                # TODO: Don't make belonging thread assumption on future usage if we need special handling
                dummpy_rt.append(RuntimeNode(
                    name='dummy',
                    start_time=None,
                    end_time=None,
                    type=EventTypes.RUNTIME,
                    tid=0,
                    device_nodes=staled_device_nodes))
                dummpy_rt[0].fill_stats()
            node_stack: List[OperatorNode] = []
            root_node = OperatorNode(
                name='CallTreeRoot',
                start_time=-sys.maxsize - 1,
                end_time=sys.maxsize,
                type=EventTypes.PYTHON,
                tid=tid,
                runtimes=zero_rt_list + dummpy_rt)  # Give the list of RuntimeNode with external_id=0 to root node.
            node_stack.append(root_node)
            for node in host_node_list:
                while True:  # break loop when the node is inserted.
                    tail_node = node_stack[-1]
                    if node.start_time < tail_node.end_time:
                        if node.end_time <= tail_node.end_time:
                            tail_node.children.append(node)
                            # node.parent_node = weakref.ref(tail_node)
                            node_stack.append(node)
                        else:
                            logger.error('Error in input data: ranges on the same thread should not intersect!'
                                         'Father:({},{},{}) Child:({},{},{})'
                                         .format(tail_node.name, tail_node.start_time, tail_node.end_time,
                                                 node.name, node.start_time, node.end_time))
                        break
                    else:
                        node_stack.pop()
            return root_node

        # Merge the consecutive calls to same function into one.
        # Just follow the same pattern in torch/autograd/profiler.py,
        # EventList._remove_dup_nodes
        # TODO: Replace recursive by for loop, in case of too deep callstack.
        def remove_dup_nodes(node: OperatorNode):
            if node.type == EventTypes.RUNTIME:
                return
            if len(node.children) == 1:
                child = node.children[0]
                if node.name == child.name and node.type == EventTypes.OPERATOR and child.type == EventTypes.OPERATOR:
                    node.children = child.children
                    node.runtimes = child.runtimes  # Keep consistent with autograd profiler.
                    remove_dup_nodes(node)  # This node may have to merge with child's child.
                    return

            for child in node.children:
                remove_dup_nodes(child)

        root_node = build_tree_relationship(host_node_list, zero_rt_list, staled_device_nodes)
        remove_dup_nodes(root_node)
        root_node.fill_stats()

        # replace the root_node start_time/end_time
        root_node.start_time = next((child.start_time for child in root_node.children
                                     if child.start_time is not None), None)
        root_node.end_time = next((child.end_time for child in reversed(root_node.children)
                                   if child.end_time is not None), None)
        return root_node

    def _get_modules(self) -> Tuple[List[ModuleNode], List[OperatorNode]]:
        """Get the ModuleNodes and backward root nodes
        If there are any ModuleNodes, the backward roots will be removed from the tree
        so that later a new BackwardNode will be replaced.
        """
        modules: List[ModuleNode] = []
        backward_nodes: Dict[OperatorNode, List[OperatorNode]] = defaultdict(list)

        def traverse_node(parent, node: OperatorNode):
            if isinstance(node, ModuleNode):
                modules.append(node)
            elif isinstance(node, ProfilerStepNode):
                for child in node.children:
                    traverse_node(node, child)
            else:
                if node.name.startswith(OpTreeBuilder.BACKWARD_ROOT_PREFIX):
                    backward_nodes[parent].append(node)
                else:
                    pass

        for root in self.tid2tree.values():
            for child in root.children:
                traverse_node(root, child)

        if modules:
            backward_nodes_flatten: List[OperatorNode] = []
            # only remove the backward nodes when the module information exist
            for p, nodes in backward_nodes.items():
                p.children = [child for child in p.children if child not in nodes]
                backward_nodes_flatten.extend(nodes)

            return modules, backward_nodes_flatten
        else:
            return None, None

    @staticmethod
    def _get_node_parents(nodes: Iterable[OperatorNode]):
        """Get the child->parent relationship for these nodes"""
        ts_to_node: Dict[int, OperatorNode] = {}
        ts_to_parent: Dict[int, OperatorNode] = {}

        def traverse_node(node: OperatorNode):
            if node.start_time not in ts_to_node:
                ts_to_node[node.start_time] = node
            for child in node.children:
                if child.start_time not in ts_to_parent:
                    ts_to_parent[child.start_time] = node
                traverse_node(child)

        for node in nodes:
            traverse_node(node)
        return ts_to_node, ts_to_parent

    @staticmethod
    def _group_backward_nodes(nodes: Iterable[OperatorNode]) -> Dict[OperatorNode, List[OperatorNode]]:
        """All nodes are backward nodes startswith autograd::engine::evaluate_function.
        If one node's name is autograd::engine::evaluate_function: torch::autograd::AccumulateGrad,
        it should be grouped with previous normal backward node. Otherwise, a new backward node should be started
        """
        grouped_bwd_nodes: List[List[OperatorNode]] = []
        for node in nodes:
            if node.name == OpTreeBuilder.BACKWARD_ACCUMULATE_GRAD:
                grouped_bwd_nodes[-1].append(node)
            else:
                grouped_bwd_nodes.append([node])

        # return the root backward node -> aggregated backward nodes array
        # if there is no any AccumulateGrad accompanied with it, then the key:value is itself.
        return {nodes[0]: nodes for nodes in grouped_bwd_nodes}

    @staticmethod
    def _get_backward_roots(fwd_bwd_map: Dict[int, int],
                            ts2parent: Dict[int, OperatorNode],
                            backward_nodes: Dict[OperatorNode, List[OperatorNode]]) -> Dict[int, List[OperatorNode]]:
        if not fwd_bwd_map:
            return None

        fwd_to_bwdroot: Dict[int, List[OperatorNode]] = {}
        for fwd, bwd in fwd_bwd_map.items():
            parent = ts2parent.get(bwd)
            while parent is not None and not parent.name.startswith(OpTreeBuilder.BACKWARD_ROOT_PREFIX):
                parent = ts2parent.get(parent.start_time)

            if parent:
                fwd_to_bwdroot[fwd] = backward_nodes.pop(parent)
            else:
                logger.warning('parent is None for', bwd)

        return fwd_to_bwdroot

    def _build_backward_module(node: ModuleNode,
                               parent: Optional[BackwardNode],
                               fwd_bwd_map: Dict[int, List[OperatorNode]],
                               result: List[BackwardNode]):
        """Construct the backward module from root (node argument) and
        insert it into result array if there is no any parent associated with it.
        """
        if not fwd_bwd_map:
            logger.warning('The forward backward map is empty. The backward construction is skipped.')
            return

        if isinstance(node, ModuleNode):
            backward_node = BackwardNode(name=node.name + '.backward', start_time=None, end_time=None,
                                         type='backward', tid=0)
            if parent is None:
                result.append(backward_node)
            else:
                parent.children.append(backward_node)
            parent = backward_node

        for child in node.children:
            if parent:
                if is_operator_node(child):
                    bwd_ops = fwd_bwd_map.get(child.start_time)
                    if bwd_ops:
                        parent.children.extend(bwd_ops)

            OpTreeBuilder._build_backward_module(child, parent, fwd_bwd_map, result)

        if isinstance(node, ModuleNode) and parent and parent.children:
            parent.fill_stats()
            parent.tid = parent.children[0].tid

    @staticmethod
    def _insert_backward_modules(root: OperatorNode, backward_modules: List[BackwardNode]):
        backward_modules.sort(key=lambda x: (x.start_time, -x.end_time))

        # each item is (parent_node, child_index) that it is visiting.
        node_stack = []
        module_index = 0
        child_index = 0
        current_node = root

        staled_modules = []

        while module_index < len(backward_modules):
            module = backward_modules[module_index]
            if current_node is None:
                # ignore all remaining modules
                staled_modules.append(module)
                module_index += 1
                continue

            if module.end_time < current_node.start_time:
                staled_modules.append(module)
                module_index += 1
                continue
            elif module.start_time > current_node.end_time:
                if node_stack:
                    # pop parent node and update the child_index accordingly.
                    current_node, child_index = node_stack.pop()
                    child_index += 1
                else:
                    # if there is not item in stack, set it to None
                    current_node = None
                continue

            while child_index < len(current_node.children):
                if module.end_time < current_node.children[child_index].start_time:
                    # if current module is before next child,
                    # we will break the search and keep the child_index not change.
                    # As the result, the module will be treated as child of 'current_node'
                    # So that next time we can continue from here.
                    # there is no any child contains the record.timestamp
                    # child_find is False at this case.
                    break
                elif module.start_time >= current_node.children[child_index].end_time:
                    child_index += 1
                else:
                    # current children contains the record
                    node_stack.append((current_node, child_index))
                    current_node = current_node.children[child_index]
                    child_index = 0

            # when code execute here, it means the current_node will be the parent of backward module
            # Add the module into current_node
            current_node.children.insert(child_index, module)
            # since the children number is increased by 1, we need increment the child_index.
            child_index += 1
            module_index += 1



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# --------------------------------------------------------------------------
from enum import IntEnum
from typing import Dict, Optional

from .. import utils

__all__ = ['EventTypes', 'create_event']

logger = utils.get_logger()


class DeviceType(IntEnum):
    CPU = 0
    CUDA = 1


class EventTypes(object):
    TRACE = 'Trace'
    OPERATOR = 'Operator'
    PROFILER_STEP = 'ProfilerStep'
    RUNTIME = 'Runtime'
    KERNEL = 'Kernel'
    MEMCPY = 'Memcpy'
    MEMSET = 'Memset'
    PYTHON = 'Python'
    MEMORY = 'Memory'
    PYTHON_FUNCTION = 'python_function'
    MODULE = 'Module'
    PL_PROFILE = 'pl_profile'
    PL_MODULE = 'pl_module'


EventTypeMap = {
    'Trace': EventTypes.TRACE,
    'cpu_op': EventTypes.OPERATOR,
    'Operator': EventTypes.OPERATOR,
    'Runtime': EventTypes.RUNTIME,
    'Kernel': EventTypes.KERNEL,
    'Memcpy': EventTypes.MEMCPY,
    'gpu_memcpy': EventTypes.MEMCPY,
    'Memset': EventTypes.MEMSET,
    'gpu_memset': EventTypes.MEMSET,
    'Python': EventTypes.PYTHON,
    'Memory': EventTypes.MEMORY,
    'python_function': EventTypes.PYTHON_FUNCTION
}


class BaseEvent(object):
    def __init__(self, type, data):
        self.type: str = type
        self.name: str = data.get('name')
        self.ts: int = data.get('ts')
        self.pid: int = data.get('pid')
        self.tid: int = data.get('tid')
        self.args: Dict = data.get('args', {})


class DurationEvent(BaseEvent):
    def __init__(self, type, data):
        super().__init__(type, data)
        self.category: str = data.get('cat', '')
        self.duration: int = data.get('dur')

        extern_id: Optional[int] = self.args.get('external id')
        if extern_id is None:
            extern_id = self.args.get('External id')
        self.external_id = extern_id
        self.correlation_id: Optional[int] = self.args.get('correlation')


class KernelEvent(DurationEvent):
    def __init__(self, type, data):
        super().__init__(type, data)
        self.occupancy = self.args.get('est. achieved occupancy %')
        self.blocks_per_sm = self.args.get('blocks per SM')
        self.grid = self.args.get('grid')
        self.block = self.args.get('block')
        self.regs_per_thread = self.args.get('registers per thread')
        self.shared_memory = self.args.get('shared memory')
        self.device_id = self.args.get('device')


class OperatorEvent(DurationEvent):
    def __init__(self, type, data):
        super().__init__(type, data)
        self.callstack = self.args.get('Call stack')
        self.input_type = self.args.get('Input type')

        shape = self.args.get('Input Dims')
        if shape is None:
            # Setting shape to '[]' other None is to align with autograd result
            shape = self.args.get('Input dims', [])
        self.input_shape = shape


class ProfilerStepEvent(OperatorEvent):
    def __init__(self, data):
        super().__init__(EventTypes.PROFILER_STEP, data)
        # torch.profiler.profile.step will invoke record_function with name like 'ProfilerStep#5'
        self.step: int = int(self.name.split('#')[1])


class MemoryEvent(BaseEvent):
    def __init__(self, type, data):
        super().__init__(type, data)
        self.scope: str = data.get('s', '')
        self.device_id: int = self.args.get('Device Id')
        dtype = self.args.get('Device Type')
        if dtype is not None:
            try:
                dtype = DeviceType(dtype)
            except ValueError:
                dtype = None

        self.device_type: DeviceType = dtype

    @property
    def addr(self):
        return self.args.get('Addr')

    @property
    def bytes(self):
        return self.args.get('Bytes', 0)

    @property
    def total_allocated(self):
        return self.args.get('Total Allocated', float('nan'))

    @property
    def total_reserved(self):
        return self.args.get('Total Reserved', float('nan'))


class PythonFunctionEvent(DurationEvent):
    def __init__(self, type, data):
        super().__init__(type, data)
        self.python_id: int = self.args.get('Python id')
        self.python_parent_id: int = self.args.get('Python parent id')


class ModuleEvent(PythonFunctionEvent):
    def __init__(self, data):
        super().__init__(EventTypes.MODULE, data)
        self.module_id: int = self.args.get('Python module id')


class PLProfileEvent(DurationEvent):
    def __init__(self, data):
        super().__init__(EventTypes.PL_PROFILE, data)
        self.name = self.name.replace('[pl][profile]', '')


class PLModuleEvent(DurationEvent):
    def __init__(self, data):
        super().__init__(EventTypes.PL_MODULE, data)
        self.module_id = 0  # just to be compatible with ModuleEvent processing
        self.name = self.name.replace('[pl][module]', '')
        # self.shape = self.name[:self.name.rfind(']')+1]
        # self.name = self.name[self.name.rfind(']')+1:]
        self.module_type = self.name[:self.name.find(': ')]
        self.name = self.name[self.name.find(': ')+2:]


def create_event(event, is_pytorch_lightning) -> Optional[BaseEvent]:
    try:
        type = event.get('ph')
        if type == 'X':
            return create_trace_event(event, is_pytorch_lightning)
        elif type == 'i' and event.get('name') == '[memory]':
            return MemoryEvent(EventTypes.MEMORY, event)
        else:
            return None
    except Exception as ex:
        logger.warning('Failed to parse profile event. Exception=%s. Event=%s', ex, event, exc_info=True)
        raise


def create_trace_event(event, is_pytorch_lightning) -> Optional[BaseEvent]:
    category = event.get('cat')
    event_type = EventTypeMap.get(category)
    if event_type == EventTypes.OPERATOR:
        name = event.get('name')
        if name and name.startswith('ProfilerStep#'):
            return ProfilerStepEvent(event)
        if is_pytorch_lightning:
            if name and name.startswith('[pl][profile]'):
                return PLProfileEvent(event)
            elif name and name.startswith('[pl][module]'):
                return PLModuleEvent(event)
        return OperatorEvent(event_type, event)
    elif event_type == EventTypes.PYTHON:
        return OperatorEvent(event_type, event)
    elif event_type == EventTypes.KERNEL:
        return KernelEvent(event_type, event)
    elif event_type == EventTypes.PYTHON_FUNCTION:
        if is_pytorch_lightning:
            return None
        args = event.get('args')
        if args and args.get('Python module id') is not None:
            return ModuleEvent(event)
        else:
            return PythonFunctionEvent(event_type, event)
    elif event_type is not None:
        return DurationEvent(event_type, event)
    else:
        return None


def create_association_events(events) -> Dict[int, int]:
    forward_map = {}
    backward_map = {}

    result = {}
    for e in events:
        ph = e.get('ph')
        id = e['id']
        ts = e['ts']
        if ph == 's':
            forward_map[id] = ts
        elif ph == 'f':
            backward_map[id] = ts

    for id, ts in forward_map.items():
        backward_ts = backward_map.get(id)
        if backward_ts is not None:
            result[ts] = backward_ts

    return result



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# --------------------------------------------------------------------------
from typing import List, Tuple

from .. import utils
from .event_parser import ProfileRole
from .range_utils import (get_ranges_sum, intersection_ranges_lists,
                          merge_ranges, subtract_ranges_lists)

logger = utils.get_logger()


class OverallParser(object):
    class Costs:
        def __init__(self, costs: List[float] = None):
            # the cost length is len(ProfileRole)
            if costs is None:
                self.costs = [0.] * len(ProfileRole)
            else:
                self.costs = costs

        @classmethod
        def create_from_statistics(cls, statistics: 'OverallParser.Statistics', total_duration: int):
            costs = [0.] * len(ProfileRole)
            for i in range(len(statistics.cost_ranges)):
                costs[i] = get_ranges_sum(statistics.cost_ranges[i])
            costs[ProfileRole.Total] = total_duration
            return cls(costs)

    class Statistics:
        def __init__(self, cost_ranges: List[List[Tuple[int, int]]]):
            if not cost_ranges:
                raise ValueError('the cost ranges is None')

            self.cost_ranges = cost_ranges

        @classmethod
        def create_from_range(cls, steps: List[Tuple[int, int]], role_ranges: List[List[Tuple[int, int]]]):
            assert len(role_ranges) == ProfileRole.Total - 1

            cost_ranges: List[List[Tuple[int, int]]] = []
            slots: List[Tuple[int, int]] = []
            for role in role_ranges:
                if slots:
                    range = intersection_ranges_lists(slots, role)
                else:
                    range = role
                    slots = merge_ranges(list(steps))
                cost_ranges.append(range)
                slots = subtract_ranges_lists(slots, range)
            # The last one is ProfileRole.Other
            cost_ranges.append(slots)

            return cls(cost_ranges)

        def intersection_with_step(self, step: Tuple[int, int]):
            cost_ranges: List[List[Tuple[int, int]]] = []
            step = [step]
            for range in self.cost_ranges:
                cost_ranges.append(intersection_ranges_lists(step, range))

            return OverallParser.Statistics(cost_ranges)

    class StepCommunicationCosts:
        def __init__(self):
            self.computation: int = 0
            self.communication: int = 0
            self.overlap: int = 0
            self.other: int = 0

    def __init__(self):
        self.steps_costs: List[OverallParser.Costs] = []
        self.avg_costs = OverallParser.Costs()
        self.communication_overlap: List[OverallParser.StepCommunicationCosts] = []

    def aggregate(self, steps: List[Tuple[int, int]], role_ranges: List[List[Tuple[int, int]]]):
        logger.debug('Overall, statistics')
        global_stats = OverallParser.Statistics.create_from_range(steps, role_ranges)
        if role_ranges[ProfileRole.Kernel]:
            comm_comp_overlap = intersection_ranges_lists(
                role_ranges[ProfileRole.Kernel], role_ranges[ProfileRole.Communication])
        else:
            comm_comp_overlap = intersection_ranges_lists(
                role_ranges[ProfileRole.CpuOp], role_ranges[ProfileRole.Communication])

        logger.debug('Overall, aggregation')
        for i, step in enumerate(steps):
            steps_stat = global_stats.intersection_with_step(step)
            self.steps_costs.append(OverallParser.Costs.create_from_statistics(steps_stat, step[1] - step[0]))
            for cost_index in range(len(self.avg_costs.costs)):
                self.avg_costs.costs[cost_index] += self.steps_costs[i].costs[cost_index]

            comm_costs = OverallParser.StepCommunicationCosts()
            comm_costs.overlap = get_ranges_sum(intersection_ranges_lists([step], comm_comp_overlap))
            if role_ranges[ProfileRole.Kernel]:
                comm_costs.computation = get_ranges_sum(
                    intersection_ranges_lists([step], role_ranges[ProfileRole.Kernel]))
            else:
                comm_costs.computation = get_ranges_sum(
                    intersection_ranges_lists([step], role_ranges[ProfileRole.CpuOp]))
            comm_costs.communication = get_ranges_sum(
                intersection_ranges_lists([step], role_ranges[ProfileRole.Communication]))
            comm_costs.other = self.steps_costs[i].costs[ProfileRole.Total] +\
                comm_costs.overlap - comm_costs.computation - comm_costs.communication
            self.communication_overlap.append(comm_costs)

        valid_steps = len(steps)
        for i in range(len(self.avg_costs.costs)):
            self.avg_costs.costs[i] /= valid_steps



def as_str_any(value):
    """Converts to `str` as `str(value)`, but use `as_str` for `bytes`.

    Args:
    value: A object that can be converted to `str`.

    Returns:
    A `str` object.
    """
    if isinstance(value, bytes):
        return as_str(value)
    else:
        return str(value)


def as_text(bytes_or_text, encoding="utf-8"):
    """Returns the given argument as a unicode string.

    Args:
    bytes_or_text: A `bytes`, `str`, or `unicode` object.
    encoding: A string indicating the charset for decoding unicode.

    Returns:
    A `str` (Python 3) object.

    Raises:
    TypeError: If `bytes_or_text` is not a binary or unicode string.
    """
    if isinstance(bytes_or_text, str):
        return bytes_or_text
    elif isinstance(bytes_or_text, bytes):
        return bytes_or_text.decode(encoding)
    else:
        raise TypeError(
            "Expected binary or unicode string, got %r" % bytes_or_text
        )


# Convert an object to a `str` in both Python 2 and 3.
as_str = as_text


def as_bytes(bytes_or_text, encoding="utf-8"):
    """Converts either bytes or unicode to `bytes`, using utf-8 encoding for
    text.

    Args:
    bytes_or_text: A `bytes`, `str`, or `unicode` object.
    encoding: A string indicating the charset for encoding unicode.

    Returns:
    A `bytes` object.

    Raises:
    TypeError: If `bytes_or_text` is not a binary or unicode string.
    """
    if isinstance(bytes_or_text, str):
        return bytes_or_text.encode(encoding)
    elif isinstance(bytes_or_text, bytes):
        return bytes_or_text
    else:
        raise TypeError(
            "Expected binary or unicode string, got %r" % (bytes_or_text,)
        )


def parse_blob_url(url):
    from urllib import parse
    url_path = parse.urlparse(url)

    parts = url_path.path.lstrip('/').split('/', 1)
    return url_path.netloc, tuple(parts)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# --------------------------------------------------------------------------
import sys
from collections import defaultdict
from typing import Callable, Dict, List

from .. import utils
from .node import DeviceNode, OperatorNode

logger = utils.get_logger()


class OperatorAgg:
    def __init__(self, op: OperatorNode):
        self.name = op.name
        self.input_shape = str(op.input_shape)  # Optional

        self.callstacks = set()  # Optional
        self.calls: int = 0
        self.host_duration: int = 0
        self.device_duration: int = 0
        self.self_host_duration: int = 0
        self.self_device_duration: int = 0
        self.tc_eligible = op.tc_eligible
        self.tc_self_duration: int = 0
        self.tc_total_duration: int = 0
        # TODO: Think about adding these avgs to UI.

    @property
    def tc_self_ratio(self) -> float:
        return self.tc_self_duration / self.self_device_duration if self.self_device_duration > 0 else 0

    @property
    def tc_total_ratio(self) -> float:
        return self.tc_total_duration / self.device_duration if self.device_duration > 0 else 0


def aggregate_ops(op_list: List[OperatorNode],
                  keys_func: List[Callable[[OperatorNode], str]]) -> List[Dict[str, OperatorAgg]]:
    def aggregate(key_to_agg: Dict[str, OperatorAgg], key: str, op: OperatorNode):
        if key not in key_to_agg:
            key_to_agg[key] = OperatorAgg(op)
        agg = key_to_agg[key]
        agg.callstacks.add(op.callstack)
        agg.calls += 1
        agg.host_duration += op.duration
        agg.device_duration += op.device_duration
        agg.self_host_duration += op.self_host_duration
        agg.self_device_duration += op.self_device_duration
        agg.tc_self_duration += op.tc_self_duration
        agg.tc_total_duration += op.tc_total_duration
        return agg

    agg_dicts: List[Dict[str, OperatorAgg]] = [{} for _ in range(len(keys_func))]
    for op in op_list:
        for i, key_func in enumerate(keys_func):
            key = key_func(op)
            aggregate(agg_dicts[i], key, op)

    return agg_dicts


class KernelAggByNameOp:
    def __init__(self, kernel: DeviceNode, op_name: str):
        self.name = kernel.name
        self.op_name = op_name
        self.grid = kernel.grid
        self.block = kernel.block
        self.regs_per_thread = kernel.regs_per_thread
        self.shared_memory = kernel.shared_memory

        self.calls: int = 0
        self.total_duration: int = 0
        self.min_duration: int = sys.maxsize
        self.max_duration: int = 0
        self.blocks_per_sm = 0.0
        self.occupancy = 0.0
        self.tc_used = kernel.tc_used
        self.op_tc_eligible = kernel.op_tc_eligible

    @property
    def avg_duration(self):
        return self.total_duration / self.calls

    @property
    def avg_blocks_per_sm(self) -> float:
        return self.blocks_per_sm / self.total_duration if self.total_duration > 0 else 0

    @property
    def avg_occupancy(self) -> float:
        return self.occupancy / self.total_duration if self.total_duration > 0 else 0


def aggregate_kernels(kernel_list: List[DeviceNode]) -> List[KernelAggByNameOp]:
    name_op_to_agg: Dict[str, KernelAggByNameOp] = {}
    for kernel in kernel_list:
        dur = kernel.end_time - kernel.start_time
        op_name = 'N/A' if kernel.op_name is None else kernel.op_name
        key = '###'.join((kernel.name, op_name,
                          str(kernel.grid), str(kernel.block),
                          str(kernel.regs_per_thread or '0'), str(kernel.shared_memory or '0')))
        if key not in name_op_to_agg:
            name_op_to_agg[key] = KernelAggByNameOp(kernel, op_name)
        agg = name_op_to_agg[key]
        agg.calls += 1
        agg.total_duration += dur
        agg.min_duration = min(agg.min_duration, dur)
        agg.max_duration = max(agg.max_duration, dur)
        agg.blocks_per_sm += float(kernel.blocks_per_sm or 0) * dur
        agg.occupancy += float(kernel.occupancy or 0) * dur

    kernel_list_groupby_name_op = list(name_op_to_agg.values())
    return kernel_list_groupby_name_op


class ModuleAggregator:

    def __init__(self):
        self.op_list_groupby_name: List[OperatorAgg] = None  # For Operator-view.
        self.op_list_groupby_name_input: List[OperatorAgg] = None  # For Operator-view.
        self.kernel_list_groupby_name_op: List[KernelAggByNameOp] = None  # For Kernel-view.
        self.stack_lists_group_by_name: Dict[str, List[OperatorAgg]] = None
        self.stack_lists_group_by_name_input: Dict[str, List[OperatorAgg]] = None
        self.ops: List[OperatorNode] = None

    def aggregate(self, tid2tree: Dict[int, OperatorNode]):
        # get the operators and kernels recursively by traverse the node tree root.
        ops: List[OperatorNode] = []
        kernels: List[DeviceNode] = []
        for root in tid2tree.values():
            root_ops, root_kernels = root.get_operator_and_kernels()
            ops.extend(root_ops)
            kernels.extend(root_kernels)

        # aggregate both kernels and operators
        self.kernel_list_groupby_name_op = aggregate_kernels(kernels)

        keys: List[Callable[[OperatorNode], str]] = [
            lambda x: x.name,
            lambda x: '###'.join((x.name, str(x.input_shape))),
            lambda x: '###'.join((x.name, str(x.callstack))),
            lambda x: '###'.join((x.name, str(x.input_shape), str(x.callstack)))
        ]
        agg_result = aggregate_ops(ops, keys)
        stack_lists_group_by_name: Dict[str, List[OperatorAgg]] = defaultdict(list)
        stack_lists_group_by_name_input: Dict[str, List[OperatorAgg]] = defaultdict(list)
        for agg in agg_result[2].values():
            assert (len(agg.callstacks) == 1)
            if list(agg.callstacks)[0]:
                stack_lists_group_by_name[agg.name].append(agg)
        for agg in agg_result[3].values():
            assert (len(agg.callstacks) == 1)
            if list(agg.callstacks)[0]:
                key = agg.name + '###' + str(agg.input_shape)
                stack_lists_group_by_name_input[key].append(agg)

        self.op_list_groupby_name = list(agg_result[0].values())
        self.op_list_groupby_name_input = list(agg_result[1].values())
        self.stack_lists_group_by_name = stack_lists_group_by_name
        self.stack_lists_group_by_name_input = stack_lists_group_by_name_input
        self.ops = ops



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# --------------------------------------------------------------------------
from collections import defaultdict
from enum import IntEnum
from typing import Dict, Iterable, List, Optional, Tuple

from .. import utils
from .node import OperatorNode, is_operator_node
from .op_agg import aggregate_ops
from .trace import DeviceType, MemoryEvent

logger = utils.get_logger()


class MemoryMetrics(IntEnum):
    SelfIncreaseSize = 0
    SelfAllocationSize = 1
    SelfAllocationCount = 2
    IncreaseSize = 3
    AllocationSize = 4
    AllocationCount = 5


class MemoryRecord:
    def __init__(self, scope: str, pid: int, tid: int, ts: int,
                 device_type: DeviceType, device_id: int,
                 address: int, bytes: int, total_allocated: float, total_reserved: float):
        self.scope = scope
        self.tid = tid
        self.pid = pid
        self.ts = ts
        self.device_type = device_type
        self.device_id = device_id
        self.addr = address
        self.bytes = bytes
        self.total_allocated = total_allocated
        self.total_reserved = total_reserved
        self.op_name: Optional[str] = None
        self.parent_op_name: Optional[str] = None

    @property
    def device_name(self):
        if self.device_type == DeviceType.CPU:
            return 'CPU'
        elif self.device_type == DeviceType.CUDA:
            return 'GPU{}'.format(self.device_id)
        else:
            return None

    @property
    def is_allocation(self):
        return self.bytes > 0

    @property
    def op_name_or_unknown(self):
        return self.op_name if self.op_name else '<unknown>'

    @classmethod
    def from_event(cls, event: MemoryEvent):
        return cls(event.scope, event.pid, event.tid, event.ts, event.device_type, event.device_id,
                   event.addr, event.bytes, event.total_allocated, event.total_reserved)

    def __repr__(self) -> str:
        return f"<{'+' if self.bytes>0 else ''}{self.bytes}B, addr: {self.addr}, ts: {self.ts}>"


class MemorySnapshot:
    def __init__(self, memory_records: Iterable[MemoryRecord],
                 op_memory_table: Dict[OperatorNode, List[MemoryRecord]],
                 processed_nodes: Dict[OperatorNode, int]) -> None:
        self.memory_records = memory_records
        self.op_memory_table = op_memory_table
        # the visited node times from parent to child
        # troubleshooting issue purpose.
        self.processed_node = processed_nodes
        self.unreached_node = defaultdict(list)

    def get_peak_memory(self) -> Dict[Tuple[DeviceType, int], int]:
        peaks = defaultdict(int)
        for r in self.memory_records:
            if r.total_allocated == r.total_allocated:  # !isnan
                peaks[(r.device_type, r.device_id)] = max(peaks[(r.device_type, r.device_id)], r.total_allocated)
        return peaks

    def get_memory_statistics(self,
                              tid2tree: Dict[int, OperatorNode],
                              start_ts=None, end_ts=None) -> Dict[str, Dict[str, List[int]]]:
        metric_length = len(MemoryMetrics)
        self_metric_length = metric_length // 2

        def dict_factory():
            return defaultdict(lambda: [0] * metric_length)

        # traverse outputs
        op_list: List[OperatorNode] = []
        # two level keys dictionary
        # first keyed by node, then keyed by device (CPU/GPU0/GPU1/etc.)
        memory_metrics_keyed_by_node: Dict[OperatorNode, Dict[str, List[int]]] = defaultdict(dict_factory)

        def traverse_node_memory(node: OperatorNode):
            if start_ts is not None and node.end_time < start_ts:
                return
            if end_ts is not None and node.start_time > end_ts:
                return

            is_op = is_operator_node(node)
            if is_op:
                op_list.append(node)

            if node not in self.processed_node:
                self.unreached_node[tid].append(node)
                # since the node has not been visited for insert memory records, just ignore all childrens
                return
            elif is_op:
                node_memory_metrics = self.get_memory_metrics(node, start_ts, end_ts)
                for device, metrics in node_memory_metrics.items():
                    # device is name of device like: CPU/GPU0
                    # metrics is an arrary [SelfIncreaseSize, SelfAllocationSize, SelfAllocationCount]
                    for i, value in enumerate(metrics):
                        memory_metrics_keyed_by_node[node][device][i] = value
                        memory_metrics_keyed_by_node[node][device][i + self_metric_length] += value
            else:
                logger.debug('node {}:{} is not operator node, will skip its self metrics processing'.format(
                    node.name, node.start_time))

            # recursive the children nodes
            for child in node.children:
                traverse_node_memory(child)
                # sum up the child metrics
                for device, metrics in memory_metrics_keyed_by_node[child].items():
                    for i in range(self_metric_length, metric_length):
                        memory_metrics_keyed_by_node[node][device][i] += metrics[i]

        for tid, root in tid2tree.items():
            for child in root.children:
                traverse_node_memory(child)

        # keyed first by device name like CPU/GPU0 etc, then keyed by operator name.
        # the value is array [items indexed by MemoryMetrics]
        memory_metrics_keyed_by_nodename: Dict[str, Dict[str, List[int]]] = defaultdict(dict_factory)
        # node: the instance, device_keyed_metrics: dictionary keyed by device name like CPU/GPU0
        for node, device_keyed_metrics in memory_metrics_keyed_by_node.items():
            if not is_operator_node(node):
                # skip the node like Optimizer.step, DataLoader, ProfilerStep#1 etc.
                continue

            for device, metrics in device_keyed_metrics.items():
                for i, metric in enumerate(metrics):
                    memory_metrics_keyed_by_nodename[device][node.name][i] += metric

        # get the op_calls dictionary from module parser result.
        op_calls: Dict[str, int] = defaultdict(int)
        agg_result = aggregate_ops(op_list, [lambda op: op.name])
        for op_name, op_agg in agg_result[0].items():
            op_calls[op_name] += op_agg.calls

        result: Dict[str, Dict[str, List[int]]] = defaultdict(defaultdict)
        for device, node_metrics in memory_metrics_keyed_by_nodename.items():
            for node, values in node_metrics.items():
                if any(values):
                    result[device][node] = values + [op_calls[node]]

        return result

    def get_memory_metrics(self, op: OperatorNode, start_ts, end_ts):
        metrics_count = len([e.name for e in MemoryMetrics if e.name.startswith('Self')])
        memory_metrics: Dict[str, List[int]] = defaultdict(lambda: [0] * metrics_count)
        for record in self.op_memory_table[op]:
            if start_ts is not None and record.ts < start_ts:
                continue
            if end_ts is not None and record.ts > end_ts:
                continue
            name = record.device_name
            if name is None:
                continue

            memory_metrics[name][MemoryMetrics.SelfIncreaseSize] += record.bytes
            if record.bytes > 0:
                memory_metrics[name][MemoryMetrics.SelfAllocationSize] += record.bytes
                memory_metrics[name][MemoryMetrics.SelfAllocationCount] += 1

        return memory_metrics


class MemoryParser:
    def __init__(self, memory_events: Iterable[MemoryEvent]):
        # statistics purpose
        self.staled_records: List[MemoryRecord] = []
        self.processed_records: List[MemoryRecord] = []
        self.memory_records: List[MemoryRecord] = [MemoryRecord.from_event(e) for e in memory_events]

    def find_memory_nodes(self, tid2tree: Dict[int, OperatorNode]) -> MemorySnapshot:
        records_by_tid: Dict[int, List[MemoryRecord]] = defaultdict(list)
        for r in self.memory_records:
            records_by_tid[r.tid].append(r)

        op_memory_table: Dict[OperatorNode, List[MemoryRecord]] = defaultdict(list)
        processed_node = defaultdict(int)

        tree_height = 0
        for tid, records in records_by_tid.items():
            if not records:
                continue

            # each item is (parent_node, child_index) that it is visiting.
            node_stack: List[Tuple[OperatorNode, int]] = []

            record_index = 0
            current_node: OperatorNode = tid2tree.get(tid)
            child_index = 0

            if current_node:
                processed_node[current_node] += 1

            while record_index < len(records):
                """In the loop, one pass will process one record. The basic logic is:
                It will search from the node that last visited since both the records and tree is ordered already
                1. it current node contains the records, then find the exactly child which just embrace it.
                2. otherwise, find the parent node and set the child_index, so that the parent node could continue from previous visited node. # noqa: E501
                3. if there is not any node contains the records, then all remaining records will be ignored.
                """
                record = records[record_index]

                if len(node_stack) > tree_height:
                    tree_height = len(node_stack)

                if current_node is None or current_node.start_time is None or current_node.end_time is None:
                    # 3. Ignore all remaining records.
                    logger.debug(
                        'could not find the node for tid %d, timestamp: %d, record index: %d, total records: %d' % (
                            record.tid, record.ts, record_index, len(records)))
                    self.staled_records.append(records[record_index])
                    record_index += 1
                    continue

                if record.ts < current_node.start_time:
                    # this should only happens for root node.
                    logger.debug('record timestamp %d is less that the start time of %s' %
                                 (record.ts, current_node.name))
                    # This record has no chance to be appended to following tree node.
                    self.staled_records.append(record)
                    record_index += 1
                    continue
                elif record.ts >= current_node.end_time:
                    # 2. pop parent node and update the child_index accordingly.
                    if len(node_stack) > 0:
                        current_node, child_index = node_stack.pop()
                        child_index += 1
                    else:
                        # if there is not item in stack, set it to None
                        current_node = None
                    continue

                # 1. find the real node embrace the record.
                # Find the node which contains the records from top to downmost.
                while child_index < len(current_node.children):
                    if record.ts < current_node.children[child_index].start_time:
                        # if current record timestamp is less than the current child's startime,
                        # we will break the search and keep the child_index not change. So that next time
                        # we can continue from here.
                        # there is no any child contains the record.timestamp
                        # child_find is False at this case.
                        break
                    elif record.ts >= current_node.children[child_index].end_time:
                        # if the record timestamp is greater than the children end time, increment to next child
                        # until find one contains the record
                        child_index += 1
                    else:
                        # current children contains the record
                        processed_node[current_node.children[child_index]] += 1

                        # push child index which will be visited, then continue the loop
                        node_stack.append((current_node, child_index))
                        current_node = current_node.children[child_index]
                        child_index = 0

                # the current_node is the one contains the record at this moment.
                if is_operator_node(current_node):
                    op_memory_table[current_node].append(record)
                    # NOTE: only allocation record can be associated with op. Because deallocation happens at the end
                    # of a tensor's lifetime which is not deterministic.
                    if record.is_allocation:
                        record.op_name = current_node.name
                        if len(node_stack) > 0:
                            record.parent_op_name = node_stack[-1][0].name
                    self.processed_records.append(record)
                else:
                    self.staled_records.append(record)

                # the record is processed
                record_index += 1

        # show summary information
        if len(self.staled_records) > 0 and len(self.memory_records) > 0:
            logger.debug('{} memory records are skipped in total {} memory records and only {} get processed'.format(
                len(self.staled_records), len(self.memory_records), len(self.processed_records)))
        if tree_height > 0:
            logger.debug('max tree height is {}'.format(tree_height))

        all_records = self.get_preprocessed_records()
        return MemorySnapshot(all_records, op_memory_table, processed_node)

    def get_preprocessed_records(self):
        memory_records = sorted(self.memory_records, key=lambda r: r.ts)

        alloc = {}  # allocation events may or may not have paired free event
        prev_ts = float('-inf')  # ensure ordered memory records is ordered
        for i, r in enumerate(memory_records):
            if r.addr is None:
                # profile json data prior to pytorch 1.10 do not have addr
                # we should ignore them
                continue
            assert prev_ts <= r.ts
            prev_ts = r.ts
            addr = r.addr
            size = r.bytes
            if size > 0:
                # Allocation event, to be matched with a Release event
                alloc[addr] = i
            else:
                # Processing a Release event
                if addr in alloc:
                    alloc_r = memory_records[alloc[addr]]
                    r.op_name = alloc_r.op_name
                    r.parent_op_name = alloc_r.parent_op_name
                    del alloc[addr]
        return memory_records
